import numpy

na = numpy.newaxis


def calc_lorentz_factor(ttheta, k:float=0.0, cthm:float = 0.91, flag_ttheta: bool=False):
    """Lorentz factor for 1D powder diffraction profile.

    k is 0 for neutrons
    k is 0.5 for characteristic x-ray radiation
    k is 0.1 for synchrotron radiation
    cthm = cos**2 (2 theta_M )
    """
    hh = 1. - k  + k * cthm * numpy.square(numpy.cos(ttheta))
    res = hh/(numpy.sin(0.5*ttheta)*numpy.sin(ttheta))
    dder = {}
    if flag_ttheta:
        dder["ttheta"] = - numpy.square(res) * \
            (0.5*numpy.cos(0.5*ttheta)*numpy.sin(ttheta) +
             numpy.sin(0.5*ttheta)*numpy.cos(ttheta)) 
    return res, dder


def func_asymmetry_f_a(z, flag_z: bool = False):
    """Function F_a(z) for asymmetry factor.
    """
    res = 2. * z * numpy.exp(-numpy.square(z))
    dder = {}
    if flag_z:
        dder["z"] = 2.*(1.-2.*numpy.square(z)) * numpy.exp(-numpy.square(z))
    return res, dder


def func_asymmetry_f_b(z, flag_z: bool = False):
    """Function F_b(z) for asymmetry factor.
    """
    f_a , dder_f_a = func_asymmetry_f_a(z, flag_z=flag_z)
    res = 2*(2*numpy.square(z)-3)*f_a
    dder = {}
    if flag_z:
        dder["z"] = 8 * z * f_a + 2*(2*numpy.square(z)-3)*dder_f_a["z"]
    return res, dder


def calc_asymmetry_factor(z, ttheta, p_1, p_2, p_3, p_4, 
        flag_z: bool = False, flag_p_1: bool = False, flag_p_2: bool = False,
        flag_p_3: bool = False, flag_p_4: bool = False):
    """Asymmetry factor for 1D powder diffraction profile.
    """
    f_a , dder_f_a = func_asymmetry_f_a(z, flag_z=flag_z)
    f_b , dder_f_b = func_asymmetry_f_b(z, flag_z=flag_z)
    
    inv_tan_th = numpy.expand_dims(1./numpy.tan(0.5*ttheta), axis=-1)
    inv_tan_tth = numpy.expand_dims(1./numpy.tan(ttheta), axis=-1)
    
    res = 1. + (p_1 * f_a + p_2 * f_b)*inv_tan_th \
         + (p_3 * f_a + p_4 * f_b)*inv_tan_tth
    dder = {}
    if flag_z:
        dder["z"] = (p_1 * dder_f_a["z"] + p_2 * dder_f_b["z"])*inv_tan_th + \
            (p_3 * dder_f_a["z"] + p_4 * dder_f_b["z"])*inv_tan_tth
    if flag_p_1:
        dder["p_1"] = f_a * inv_tan_th
    if flag_p_2:
        dder["p_2"] = f_b * inv_tan_th
    if flag_p_3:
        dder["p_3"] = f_a * inv_tan_tth
    if flag_p_4:
        dder["p_4"] = f_b * inv_tan_tth

    return res, dder


def func_gauss_by_h_pv(z, h_pv, flag_z: bool = False, flag_h_pv: bool = False):
    """Gauss function as function of h_pv
    """
    inv_h_pv = 1./h_pv 
    inv_h_pv_sq = numpy.square(inv_h_pv)
    z_deg = z * 180./numpy.pi
    z_deg_sq = numpy.square(z_deg)
    a_g = 2.*inv_h_pv*numpy.sqrt(numpy.log(2)/numpy.pi)
    b_g = 4.*numpy.log(2)*inv_h_pv_sq
    res = numpy.expand_dims(a_g, axis=-1) * numpy.exp(-numpy.expand_dims(b_g, axis=-1) * z_deg_sq) # numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
    dder = {}
    if  flag_z:
        dder["z"] = -2.*z_deg*res * 180./numpy.pi
    if flag_h_pv:
        dder["h_pv"] = (-1. + 2.* numpy.expand_dims(b_g, axis=-1)*z_deg_sq)*res*numpy.expand_dims(inv_h_pv, axis=-1)
    return res, dder


def func_lorentz_by_h_pv(z, h_pv, flag_z: bool = False, flag_h_pv: bool = False):
    """Gauss function as function of h_pv
    """
    inv_h_pv = 1./h_pv
    inv_h_pv_sq = numpy.square(inv_h_pv)
    z_deg = z * 180./numpy.pi
    c_a = 2./numpy.pi
    a_l = c_a * inv_h_pv
    b_l = 4.*inv_h_pv_sq
    z_deg_sq = numpy.square(z_deg)
    res = numpy.expand_dims(a_l, axis=-1) /(1+ numpy.expand_dims(b_l, axis=-1) * z_deg_sq)
    dder = {}
    if  flag_z:
        dder["z"] = -2.*z_deg*numpy.expand_dims(b_l,axis=-1)*res/(1.+numpy.expand_dims(b_l, axis=-1)*z_deg_sq) * 180./numpy.pi
    if flag_h_pv:
        dder["h_pv"] = (c_a * (numpy.expand_dims(h_pv, axis=-1) + 4*z_deg_sq) - \
            c_a * numpy.expand_dims(h_pv, axis=-1))/numpy.square(numpy.expand_dims(h_pv, axis=-1) + 4*z_deg_sq)
    return res, dder


def calc_h_g(u, v, w, i_g, ttheta,
        flag_u: bool = False, flag_v: bool = False, flag_w: bool = False, flag_i_g: bool = False):
    """Calculate H_G. 
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """
    tan_th  = numpy.tan(0.5*ttheta)
    tan_th_sq = numpy.square(tan_th)
    inv_c_th_sq = 1./numpy.square(numpy.cos(0.5*ttheta))
    hh = u*tan_th_sq + v*tan_th + w + i_g * inv_c_th_sq
    hh[ hh<numpy.finfo(float).eps ] = numpy.finfo(float).eps # protection against negative variables
    res = numpy.sqrt(hh)
    dder = {}
    if flag_u:
        dder["u"] = 0.5*tan_th_sq/res
    if flag_v:
        dder["v"] = 0.5*tan_th/res
    if flag_w:
        dder["w"] = 0.5/res
    if flag_i_g:
        dder["i_g"] = 0.5*inv_c_th_sq/res
    return res, dder


def calc_h_l(x, y, ttheta, flag_x: bool = False, flag_y: bool = False):
    """Calculate H_L. 
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """
    tan_th  = numpy.tan(0.5*ttheta)
    inv_c_th = 1./numpy.cos(0.5*ttheta)
    res = x * tan_th + y*inv_c_th
    dder = {}
    if flag_x:
        dder["x"] = tan_th
    if flag_y:
        dder["y"] = inv_c_th
    return res, dder


def calc_h_pv(h_g, h_l, flag_h_g: bool = False, flag_h_l: bool = False):
    """Calculate H_pV. 
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """
    h_g_2, h_l_2 = numpy.square(h_g), numpy.square(h_l)
    h_g_3, h_l_3 = h_g_2*h_g, h_l_2*h_l
    h_g_4, h_l_4 = numpy.square(h_g_2), numpy.square(h_l_2)
    h_g_5, h_l_5 = h_g_4*h_g, h_l_4*h_l

    c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
    hh = h_g_5 + c_2*h_g_4*h_l + c_3*h_g_3*h_l_2 + c_4*h_g_2*h_l_3 + c_5*h_g*h_l_4 + h_l_5
    res = numpy.power(hh, 0.2)

    dder = {}
    if flag_h_g or flag_h_l:
        c_help = -0.2*numpy.power(hh, -0.8)
    if flag_h_g:
        dder["h_g"] = c_help * (5*h_g_4 + 4*c_2*h_g_3*h_l + 3*c_3*h_g_2*h_l_2 + 2*c_4*h_g*h_l_3 + c_5*h_l_4)
    if flag_h_l:
        dder["h_l"] = c_help * (c_2*h_g_4 + 2*c_3*h_g_3*h_l + 3*c_4*h_g_2*h_l_2 + 4*c_5*h_g*h_l_3 + 5*h_l_4)
    return res, dder


def calc_eta(h_l, h_pv, flag_h_l: bool = False, flag_h_pv: bool = False):
    """Calculate H_pV. 
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """
    c_1 = 1.36603
    c_2 = -0.47719
    c_3 = 0.11116

    rel = h_l/h_pv
    rel_sq = numpy.square(rel)
    res = c_1*rel + c_2*rel_sq  + c_3*rel_sq*rel

    dder = {}
    if flag_h_pv or flag_h_l:
        c_help = c_1 + 2*c_2*rel  + 3*c_3*rel_sq
    if flag_h_pv:
        dder["h_pv"] = -1.*c_help * rel / h_pv 
    if flag_h_l:
        dder["h_l"] = c_help/h_pv
    return res, dder


def calc_profile_pseudo_voight(ttheta, ttheta_hkl, u, v, w, i_g, x, y,
        p_1, p_2, p_3, p_4, 
        flag_ttheta: bool=False, 
        flag_ttheta_hkl: bool=False, flag_u: bool=False,
        flag_v: bool=False, flag_w: bool=False, flag_i_g: bool=False, flag_x: bool=False,
        flag_y: bool=False, flag_p_1: bool = False, flag_p_2: bool = False,
        flag_p_3: bool = False, flag_p_4: bool = False):
    """Calculate profile as psevdo-Voight function defined by parameters.
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """
    delta_angle = (ttheta[:, na] - ttheta_hkl[na, :])
    flag_delta_angle = flag_ttheta_hkl or flag_ttheta

    flag_h_g = flag_u or flag_v or flag_w or flag_i_g
    h_g, dder_h_g = calc_h_g(
        u,v, w, i_g, ttheta,
        flag_u=flag_u, flag_v=flag_v, flag_w=flag_w, flag_i_g=flag_i_g)

    flag_h_l = flag_x or flag_y
    h_l, dder_h_l = calc_h_l(
        x, y, ttheta, flag_x=flag_x, flag_y=flag_y)

    flag_h_pv = flag_h_g or flag_h_l
    h_pv, dder_h_pv = calc_h_pv(h_g, h_l, flag_h_g=flag_h_g, flag_h_l=flag_h_l)

    eta, dder_eta = calc_eta(h_l, h_pv, flag_h_l=flag_h_l, flag_h_pv=flag_h_pv)

    z = (delta_angle*180./numpy.pi)/numpy.expand_dims(h_pv, axis=1)
    flag_z = flag_h_pv or flag_ttheta_hkl or flag_ttheta
    af, dder_af  = calc_asymmetry_factor(z, ttheta, p_1, p_2, p_3, p_4, 
        flag_z = flag_z, flag_p_1 =  flag_p_1, flag_p_2 =  flag_p_2,
        flag_p_3 = flag_p_3, flag_p_4 = flag_p_4)
    
    profile_l, dder_profile_l = func_lorentz_by_h_pv(
        delta_angle, h_pv, flag_z=flag_delta_angle, flag_h_pv=flag_h_pv)
    profile_g, dder_profile_l = func_gauss_by_h_pv(
        delta_angle, h_pv, flag_z=flag_delta_angle, flag_h_pv=flag_h_pv)

    res = (numpy.expand_dims(eta, axis=1) * profile_l + numpy.expand_dims((1.-eta), axis=1)*profile_g)*af
    dder = {}
    return res, dder


def calc_gamma_nu_by_ttheta_phi(ttheta, phi, flag_ttheta: bool = False, flag_phi: bool = False):
    """See the documentation module "Powder diffraction at constant wavelength".
    """
    s_tth, c_tth = numpy.sin(ttheta), numpy.cos(ttheta)
    s_phi, c_phi = numpy.sin(phi), numpy.cos(phi)

    gamma = numpy.arctan2(s_tth*c_phi, c_tth)
    nu = numpy.arcsin(s_tth*s_phi)

    dder_gamma = {}
    dder_nu = {}
    return gamma, nu, dder_gamma, dder_nu


def calc_ttheta_phi_by_gamma_nu(gamma, nu, flag_gamma: bool = False, flag_nu: bool = False):
    """See the documentation module "Powder diffraction at constant wavelength".
    """
    s_ga, c_ga = numpy.sin(gamma), numpy.cos(gamma)
    s_nu, c_nu = numpy.sin(nu), numpy.cos(nu)

    ttheta = numpy.arccos(c_nu*c_ga)
    phi = numpy.arctan2(s_nu, s_ga*c_nu)

    dder_ttheta = {}
    dder_phi = {}
    return ttheta, phi, dder_ttheta, dder_phi


def calc_profile_pseudo_voight_2d(ttheta, phi,
        ttheta_hkl, u, v, w, i_g, x, y,
        p_1, p_2, p_3, p_4, p_phi,
        flag_ttheta: bool=False, 
        flag_ttheta_hkl: bool=False, flag_u: bool=False,
        flag_v: bool=False, flag_w: bool=False, flag_i_g: bool=False, flag_x: bool=False,
        flag_y: bool=False, flag_p_1: bool = False, flag_p_2: bool = False,
        flag_p_3: bool = False, flag_p_4: bool = False, flag_p_phi: bool = False):
    """Calculate profile as psevdo-Voight function defined by parameters.
    For more documentation see documentation module "Powder diffraction at constant wavelenght".
    """

    delta_ttheta = ttheta[:, :, na] - ttheta_hkl[na, na, :]
    flag_delta_angle = flag_ttheta_hkl or flag_ttheta

    flag_h_g = flag_u or flag_v or flag_w or flag_i_g
    h_g, dder_h_g = calc_h_g(u,v, w, i_g, ttheta, flag_u=flag_u, flag_v=flag_v, flag_w=flag_w, flag_i_g=flag_i_g)

    flag_h_l = flag_x or flag_y
    h_l, dder_h_l = calc_h_l(x, y, ttheta, flag_x=flag_x, flag_y=flag_y)

    flag_h_pv = flag_h_g or flag_h_l
    h_pv, dder_h_pv = calc_h_pv(h_g, h_l, flag_h_g=flag_h_g, flag_h_l=flag_h_l)

    eta, dder_eta = calc_eta(h_l, h_pv, flag_h_l=flag_h_l, flag_h_pv=flag_h_pv)


    coeff_phi = calc_coeff_phi(phi, p_phi, flag_p_phi=flag_p_phi)[0]
    h_pv_phi = h_pv*coeff_phi
    flag_h_pv_phi = flag_h_pv or flag_p_phi

    z = (delta_ttheta*180./numpy.pi)/numpy.expand_dims(h_pv_phi, axis=-1)
    flag_z = flag_h_pv_phi or flag_ttheta_hkl or flag_ttheta
    af, dder_af  = calc_asymmetry_factor(z, ttheta, p_1, p_2, p_3, p_4, 
        flag_z = flag_z, flag_p_1 =  flag_p_1, flag_p_2 =  flag_p_2,
        flag_p_3 = flag_p_3, flag_p_4 = flag_p_4)
    
    profile_l, dder_profile_l = func_lorentz_by_h_pv(delta_ttheta, h_pv_phi, flag_z=flag_delta_angle, flag_h_pv=flag_h_pv_phi)
    profile_g, dder_profile_l = func_gauss_by_h_pv(delta_ttheta, h_pv_phi, flag_z=flag_delta_angle, flag_h_pv=flag_h_pv_phi)

    res = (numpy.expand_dims(eta, axis=-1) * profile_l + numpy.expand_dims((1.-eta), axis=-1)*profile_g)*af
    dder = {}
    return res, dder


def calc_coeff_phi(phi, p_phi, flag_p_phi: bool=False):
    res= 0.5*(p_phi+1.)-numpy.cos(phi)*(p_phi-1.)*0.5
    dder = {}
    if flag_p_phi:
        dder["p_phi"] = 0.5-0.5*numpy.cos(phi)
    return res, dder


"""Function to operate with mcif.

    - read_magnetic_data
    - calc_fract_by_sym_elem
    - calc_moment_by_sym_elem
    - get_sym_elem_by_ops
    - find_i_for_nlabel_bns
    - get_sym_elems_magn_centering_for_bns
    - get_str_for_sym_elem
    - calc_full_sym_elems
    - rational_sum
    - calc_multiplicity
"""

import os
import numpy

from typing import Tuple

from cryspy.A_functions_base.function_1_matrices import \
    calc_determinant_matrix_ij

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_sthovl_by_hkl_abc_angles

# for the ith nonhexagonal point operator:
# point_op_label(i): point operator symbol (from Litvin)
point_op_label = numpy.zeros(shape=(48, ), dtype='<8U')
# point_op_xyz(i): point operator in x,y,z notation
point_op_xyz = numpy.zeros(shape=(48, ), dtype='<10U')
# point_op_matrix(i): point operator matrix
point_op_matrix = numpy.zeros(shape=(3, 3, 48, ), dtype=int)
# for the ith hexagonal point operator:
# point_op_hex_label(i): point operator symbol (from Litvin)
point_op_hex_label = numpy.zeros(shape=(24, ), dtype='<8U')
# point_op_hex_xyz(i): point operator in x,y,z notation
point_op_hex_xyz = numpy.zeros(shape=(24, ), dtype='<10U')
# point_op_hex_matrix(i): point operator matrix
point_op_hex_matrix = numpy.zeros(shape=(3, 3, 24, ), dtype=int)

# number of magnetic space groups
magcount = 1651
# for the ith magnetic space group
# nlabel_bns(i): numerical label in BNS setting
nlabel_bns = numpy.zeros(shape=(magcount, ), dtype='<12U')
# nlabel_parts_bns(j,i): jth part of nlabel_bns
nlabelparts_bns = numpy.zeros(shape=(2, magcount, ), dtype=int)
# label_bns(i): group symbol
spacegroup_label_bns = numpy.zeros(shape=(magcount, ), dtype='<14U')
# nlabel_og(i): numerical label in OG setting
nlabel_og = numpy.zeros(shape=(magcount, ), dtype='<12U')
# nlabel_parts_og(j,i): jth part of nlabel_og
nlabelparts_og = numpy.zeros(shape=(3, magcount, ), dtype=int)
# label_og(i): group symbol
spacegroup_label_og = numpy.zeros(shape=(magcount, ), dtype='<14U')
# magtype(i): type of magnetic space group (1-4)
magtype = numpy.zeros(shape=(magcount, ), dtype=int)
# BNS-OG transformation (if type-4)
# bnsog_point_op(j,k,i): 3x3 point operator part of transformation
bnsog_point_op = numpy.zeros(shape=(3, 3, magcount, ), dtype=int)
# bnsog_origin(j,i): translation part of transformation
# bnsog_point_origin(i): common denominator
bnsog_origin = numpy.zeros(shape=(3, magcount, ), dtype=int)
bnsog_origin_denom = numpy.zeros(shape=(magcount, ), dtype=int)
# iops_count(i): number of point operators
ops_count = numpy.zeros(shape=(magcount, ), dtype=int)
# wyckoff_count(i): number of wyckoff sites
wyckoff_site_count = numpy.zeros(shape=(magcount, ), dtype=int)
# wyckoff_pos_count(j,i): number of positions in jth wyckoff site
wyckoff_pos_count = numpy.zeros(shape=(27, magcount, ), dtype=int)
# wyckoff_mult(j,i): multiplicity for jth wyckoff site
wyckoff_mult = numpy.zeros(shape=(27, magcount, ), dtype=int)
# wyckoff_label(j,i): symbol (a,b,c,...,z,alpha) for jth wyckoff site
wyckoff_label = numpy.zeros(shape=(27, magcount, ), dtype="<5U")
# for BNS setting
# lattice_bns_vectors_count(i): number of lattice vectors defining the lattice
lattice_bns_vectors_count = numpy.zeros(shape=(magcount, ), dtype=int)
# lattice_bns_vectors(k,j,i): kth component of the jth lattice vector
# lattice_bns_vectors_denom(j,i): common denominator
lattice_bns_vectors = numpy.zeros(shape=(3, 6, magcount, ), dtype=int)
lattice_bns_vectors_denom = numpy.zeros(shape=(6, magcount, ), dtype=int)
# for jth operator
# ops_bns_point_op(j,i): point operator part
ops_bns_point_op = numpy.zeros(shape=(96, magcount, ), dtype=int)
# ops_bns_trans(k,j,i): kth component of translation part
# ops_bns_trans_denom(j,i): common denominator
ops_bns_trans = numpy.zeros(shape=(3, 96, magcount, ), dtype=int)
ops_bns_trans_denom = numpy.zeros(shape=(96, magcount, ), dtype=int)
# ops_bns_timeinv(j,i): 1=no time inversion, -1=time inversion
ops_bns_timeinv = numpy.zeros(shape=(96, magcount, ), dtype=int)
# for jth wyckoff site
# wyckoff_bns_fract(k,j,i): kth component of fractional part of wyckoff
# position
# wyckoff_bns_fract_denom(j,i): common denominator
wyckoff_bns_fract = numpy.zeros(shape=(3, 96, 27, magcount, ), dtype=int)
wyckoff_bns_fract_denom = numpy.zeros(shape=(96, 27, magcount, ), dtype=int)
# wyckoff_bns_xyz(m,k,j,i): mth component to coeffcient of kth paramater
# (x,y,z)
wyckoff_bns_xyz = numpy.zeros(shape=(3, 3, 96, 27, magcount, ), dtype=int)
# wyckoff_bns_mag(m,k,j,i): mth component to coeffcient of kth magnetic
# paramater (mx,my,mz)
wyckoff_bns_mag = numpy.zeros(shape=(3, 3, 96, 27, magcount, ), dtype=int)

# for OG setting (for type-4 groups)
# lattice_og_vectors_count(i): number of lattice vectors defining the lattice
lattice_og_vectors_count = numpy.zeros(shape=(magcount, ), dtype=int)
# lattice_og_vectors(k,j,i): kth component of the jth lattice vector
# lattice_og_vectors_denom(j,i): common denominator
lattice_og_vectors = numpy.zeros(shape=(3, 6, magcount, ), dtype=int)
lattice_og_vectors_denom = numpy.zeros(shape=(6, magcount, ), dtype=int)
# for jth operator
# ops_og_point_op(j,i): point operator part
ops_og_point_op = numpy.zeros(shape=(96, magcount, ), dtype=int)
# ops_og_trans(k,j,i): kth component of translation part
# ops_og_trans_denom(j,i): common denominator
ops_og_trans = numpy.zeros(shape=(3, 96, magcount, ), dtype=int)
ops_og_trans_denom = numpy.zeros(shape=(96, magcount, ), dtype=int)
# ops_og_timeinv(j,i): 1=no time inversion, -1=time inversion
ops_og_timeinv = numpy.zeros(shape=(96, magcount, ), dtype=int)
# for jth wyckoff site
# wyckoff_og_fract(k,j,i): kth component of fractional part of wyckoff position
# wyckoff_og_fract_denom(j,i): common denominator
wyckoff_og_fract = numpy.zeros(shape=(3, 96, 27, magcount, ), dtype=int)
wyckoff_og_fract_denom = numpy.zeros(shape=(96, 27, magcount, ), dtype=int)
# wyckoff_og_xyz(m,k,j,i): mth component to coeffcient of kth paramater (x,y,z)
wyckoff_og_xyz = numpy.zeros(shape=(3, 3, 96, 27, magcount, ), dtype=int)
# wyckoff_og_mag(m,k,j,i): mth component to coeffcient of kth magnetic
# paramater (mx,my,mz)
wyckoff_og_mag = numpy.zeros(shape=(3, 3, 96, 27, magcount, ), dtype=int)

F_MAG_DATA = os.path.join(os.path.dirname(__file__), "magnetic_data.txt")

D_CENTRING_TYPE_SHIFT = {
    "P": numpy.array([[0, 0, 0, 1], ], dtype=int).transpose(),
    "A": numpy.array([[0, 0, 0, 1], [0, 1, 1, 2]], dtype=int).transpose(),    
    "B": numpy.array([[0, 0, 0, 1], [1, 0, 1, 2]], dtype=int).transpose(),    
    "C": numpy.array([[0, 0, 0, 1], [1, 1, 0, 2]], dtype=int).transpose(),    
    "F": numpy.array([[0, 0, 0, 1], [0, 1, 1, 2], [1, 0, 1, 2], [1, 1, 0, 2]],
                     dtype=int).transpose(),    
    "I": numpy.array([[0, 0, 0, 1], [1, 1, 1, 2]], dtype=int).transpose(),    
    "R": numpy.array([[0, 0, 0, 1], [2, 1, 1, 3], [1, 2, 2, 3]],
                     dtype=int).transpose(),    
    "Rrev": numpy.array([[0, 0, 0, 1], [1, 2, 1, 3], [2, 1, 2, 3]],
                        dtype=int).transpose(),    
    "H": numpy.array([[0, 0, 0, 1], [2, 1, 0, 3], [1, 2, 0, 3]],
                     dtype=int).transpose(),    
        }

def read_magnetic_data():
    """Read magnetic data."""
    with open(F_MAG_DATA, "r") as fid:
        l_cont = fid.readlines()
    # read nonhexangonal point operators
    i_line = 0
    for i in range(48):
        l_h = l_cont[i_line].split()
        i_line += 1
        n = int(l_h[0])
        point_op_label[i] = l_h[1]
        point_op_xyz[i] = l_h[2]
        point_op_matrix[:, :, i] = numpy.array(l_h[3:3+9], dtype=int).reshape(
            3, 3)
        if n != (i+1):
            break
    # read hexangonal point operators
    for i in range(24):
        l_h = l_cont[i_line].split()
        i_line += 1
        n = int(l_h[0])
        point_op_hex_label[i] = l_h[1]
        point_op_hex_xyz[i] = l_h[2]
        point_op_hex_matrix[:, :, i] = numpy.array(l_h[3:3+9], dtype=int
                                                   ).reshape(3, 3)
        if n != (i+1):
            break

    for i in range(magcount):
        l_h = l_cont[i_line].split()
        i_line += 1
        nlabelparts_bns[:, i] = numpy.array(l_h[0:2], dtype=int)
        nlabel_bns[i] = l_h[2].strip("\"")
        spacegroup_label_bns[i] = l_h[3].strip("\"")
        nlabelparts_og[:, i] = numpy.array(l_h[4:7], dtype=int)
        nlabel_og[i] = l_h[7]
        spacegroup_label_og[i] = l_h[8].strip("\"")

        l_h = l_cont[i_line].split()
        i_line += 1
        magtype[i] = l_h[0]

        if magtype[i] == 4:
            l_h = l_cont[i_line].split()
            i_line += 1
            bnsog_point_op[:, :, i] = numpy.transpose(
                numpy.array(l_h[0:9], dtype=int).reshape(3, 3))
            bnsog_origin[:, i] = numpy.array(l_h[9:9+3], dtype=int)
            bnsog_origin_denom[i] = l_h[12]

        l_h = l_cont[i_line].split()
        i_line += 1
        ops_count[i] = l_h[0]

        for j in range(ops_count[i]):
            if j % 4 == 0:
                l_h = l_cont[i_line].split()
                i_line += 1
            ops_bns_point_op[j, i], ops_bns_trans[0, j, i], \
                ops_bns_trans[1, j, i], ops_bns_trans[2, j, i], \
                ops_bns_trans_denom[j, i], ops_bns_timeinv[j, i] = \
                l_h[0+6*(j % 4)], l_h[1+6*(j % 4)], l_h[2+6*(j % 4)],\
                l_h[3+6*(j % 4)], l_h[4+6*(j % 4)], l_h[5+6*(j % 4)]

        l_h = l_cont[i_line].split()
        i_line += 1
        lattice_bns_vectors_count[i] = l_h[0]

        l_h = l_cont[i_line].split()
        i_line += 1
        for j in range(lattice_bns_vectors_count[i]):
            lattice_bns_vectors[0, j, i], lattice_bns_vectors[1, j, i], \
                lattice_bns_vectors[2, j, i], \
                lattice_bns_vectors_denom[j, i] = \
                l_h[0+4*j], l_h[1+4*j], l_h[2+4*j], l_h[3+4*j]

        l_h = l_cont[i_line].split()
        i_line += 1
        wyckoff_site_count[i] = l_h[0]

        for j in range(wyckoff_site_count[i]):
            l_h = l_cont[i_line].split()
            i_line += 1
            wyckoff_pos_count[j, i],  wyckoff_mult[j, i], \
                wyckoff_label[j, i] = l_h[0], l_h[1], l_h[2][1:]

            for k in range(wyckoff_pos_count[j, i]):
                l_h = l_cont[i_line].split()
                i_line += 1
                wyckoff_bns_fract[:, k, j, i] = numpy.array(l_h[0:3],
                                                            dtype=int)
                wyckoff_bns_fract_denom[k, j, i] = l_h[3]
                wyckoff_bns_xyz[:, :, k, j, i] = numpy.transpose(
                    numpy.array(l_h[4:13], dtype=int).reshape(3, 3))
                wyckoff_bns_mag[:, :, k, j, i] = numpy.transpose(
                    numpy.array(l_h[13:22], dtype=int).reshape(3, 3))

        if magtype[i] == 4:
            l_h = l_cont[i_line].split()
            i_line += 1
            ops_count[i] = l_h[0]

            for j in range(ops_count[i]):
                if j % 4 == 0:
                    l_h = l_cont[i_line].split()
                    i_line += 1
                ops_og_point_op[j, i], ops_og_trans[0, j, i], \
                    ops_og_trans[1, j, i], ops_og_trans[2, j, i], \
                    ops_og_trans_denom[j, i], ops_og_timeinv[j, i] = \
                    l_h[0+6*(j % 4)], l_h[1+6*(j % 4)], l_h[2+6*(j % 4)], \
                    l_h[3+6*(j % 4)], l_h[4+6*(j % 4)], l_h[5+6*(j % 4)]

            l_h = l_cont[i_line].split()
            i_line += 1
            lattice_og_vectors_count[i] = l_h[0]

            l_h = l_cont[i_line].split()
            i_line += 1
            for j in range(lattice_og_vectors_count[i]):
                lattice_og_vectors[:, j, i] = numpy.array(l_h[0+4*j:3+4*j],
                                                          dtype=int)
                lattice_og_vectors_denom[j, i] = l_h[3+4*j]

            l_h = l_cont[i_line].split()
            i_line += 1
            wyckoff_site_count[i] = l_h[0]

            for j in range(wyckoff_site_count[i]):
                l_h = l_cont[i_line].split()
                i_line += 1
                wyckoff_pos_count[j, i], wyckoff_mult[j, i] = l_h[0], l_h[1]
                wyckoff_label[j, i] = l_h[2][1:]

                for k in range(wyckoff_pos_count[j, i]):
                    l_h = l_cont[i_line].split()
                    i_line += 1
                    wyckoff_og_fract[:, k, j, i] = numpy.array(l_h[0:3],
                                                               dtype=int)
                    wyckoff_og_fract_denom[k, j, i] = l_h[3]
                    wyckoff_og_xyz[:, :, k, j, i] = numpy.transpose(
                        numpy.array(l_h[4:4+9], dtype=int).reshape(3, 3))
                    wyckoff_og_mag[:, :, k, j, i] = numpy.transpose(
                        numpy.array(l_h[13:13+9], dtype=int).reshape(3, 3))


def calc_fract_by_sym_elem(sym_elems, fract):
    """Calculate fraction of atom for given atom and symmetry element."""
    fract_np = numpy.array(fract, dtype=float)

    if len(fract_np.shape) == 1:
        fract_np.resize((3,1))

    f_x, f_y, f_z = fract_np[0], fract_np[1], fract_np[2]
    b_x, b_y, b_z = sym_elems[0], sym_elems[1], sym_elems[2]
    denom = sym_elems[3]
    r_11, r_12, r_13 = sym_elems[4], sym_elems[5], sym_elems[6]
    r_21, r_22, r_23 = sym_elems[7], sym_elems[8], sym_elems[9]
    r_31, r_32, r_33 = sym_elems[10], sym_elems[11], sym_elems[12]

    n_a = numpy.newaxis
    at_x_new = numpy.mod(r_11[:, n_a] * f_x[n_a, :] + r_12[:, n_a] *
                         f_y[n_a, :] + r_13[:, n_a] * f_z[n_a, :] +
                         (b_x/denom)[:, n_a], 1.)
    at_y_new = numpy.mod(r_21[:, n_a] * f_x[n_a, :] + r_22[:, n_a] *
                         f_y[n_a, :] + r_23[:, n_a] * f_z[n_a, :] +
                         (b_y/denom)[:, n_a], 1.)
    at_z_new = numpy.mod(r_31[:, n_a] * f_x[n_a, :] + r_32[:, n_a] *
                         f_y[n_a, :] + r_33[:, n_a] * f_z[n_a, :] +
                         (b_z/denom)[:, n_a], 1.)
    at_new = numpy.array([at_x_new, at_y_new, at_z_new], dtype=float)
    return at_new


def calc_moment_by_sym_elem(sym_elems, moment):
    """Calculate fraction of atom for given atom and symmetry element."""
    moment_np = numpy.array(moment, dtype=float)

    if len(moment_np.shape) == 1:
        moment_np.resize((3,1))

    m_x, m_y, m_z = moment_np[0], moment_np[1], moment_np[2]

    r_11, r_12, r_13 = sym_elems[4], sym_elems[5], sym_elems[6]
    r_21, r_22, r_23 = sym_elems[7], sym_elems[8], sym_elems[9]
    r_31, r_32, r_33 = sym_elems[10], sym_elems[11], sym_elems[12]

    theta = sym_elems[13]

    n_a = numpy.newaxis
    # FIXME: multiplication on det(R) is probably incorrect for hexagonal
    # structures
    det_r = calc_determinant_matrix_ij((r_11, r_12, r_13, r_21, r_22, r_23,
                                        r_31, r_32, r_33))

    m_x_new = (r_11[:, n_a] * m_x[n_a, :] + r_12[:, n_a] * m_y[n_a, :] +
               r_13[:, n_a] * m_z[n_a, :]) * (det_r*theta)[:, n_a]
    m_y_new = (r_21[:, n_a] * m_x[n_a, :] + r_22[:, n_a] * m_y[n_a, :] +
               r_23[:, n_a] * m_z[n_a, :]) * (det_r*theta)[:, n_a]
    m_z_new = (r_31[:, n_a] * m_x[n_a, :] + r_32[:, n_a] * m_y[n_a, :] +
               r_33[:, n_a] * m_z[n_a, :]) * (det_r*theta)[:, n_a]
    m_new = numpy.array([m_x_new, m_y_new, m_z_new], dtype=float)
    return m_new


def get_sym_elem_by_ops(ops, flag_non_hexagonal: bool = True):
    """Get symmetry element by operator.

    oops = [point_group_number, f_x, f_y, f_z, denom, time_inversion]

    flag_non_hexagonal is True for nonhexagonal point group number.
    """
    try:
        size_ops = len(ops[0])  # if ops is list of numpy arrays
    except TypeError:
        size_ops = 1  # if ops is list of numbers

    t_i = ops[5]  # time inversion

    if flag_non_hexagonal:
        val = numpy.reshape(point_op_matrix[:, :, ops[0]-1], (9, size_ops))
    else:
        val = numpy.reshape(point_op_hex_matrix[:, :, ops[0]-1], (9, size_ops))
    res = numpy.array([
        ops[1], ops[2], ops[3], ops[4], val[0], val[1], val[2], val[3], val[4],
        val[5], val[6], val[7], val[8], t_i*val[0], t_i*val[1], t_i*val[2],
        t_i*val[3], t_i*val[4], t_i*val[5], t_i*val[6], t_i*val[7],
        t_i*val[8]], dtype=int)
    return res


def find_i_for_nlabel_bns(part_1: int, part_2: int):
    flag_1 = nlabelparts_bns[0, :] == part_1
    flag_2 = nlabelparts_bns[1, :] == part_2
    flag_3 = flag_1*flag_2
    return int(numpy.argwhere(flag_3))



def get_sym_elems_magn_centering_for_bns(part_1: int, part_2: int):
    """Get symmetry elements for bns space group."""
    i = find_i_for_nlabel_bns(part_1, part_2)
    j = ops_count[i]
    ops = [ops_bns_point_op[:j, i], ops_bns_trans[0, :j, i],
           ops_bns_trans[1, :j, i], ops_bns_trans[2, :j, i],
           ops_bns_trans_denom[:j, i], ops_bns_timeinv[:j, i]]

    flag_non_hexagonal = True  # FIXME
    sym_elems = get_sym_elem_by_ops(ops, flag_non_hexagonal=flag_non_hexagonal)

    label = spacegroup_label_bns[i]
    n_shift = D_CENTRING_TYPE_SHIFT[label[0]]
    new_shape = (18, )+n_shift.shape[1:]
    xyz_sym = numpy.zeros(shape=new_shape, dtype=int)
    xyz_sym[0], xyz_sym[4], xyz_sym[8] = 1, 1, 1
    xyz_sym[9], xyz_sym[13], xyz_sym[17] = 1, 1, 1
    magn_centering = numpy.vstack([n_shift, xyz_sym])
    return sym_elems, magn_centering


def get_str_for_sym_elem(sym_elem, labels=("x", "y", "z", "mx", "my", "mz")):
    """Get str for sym element."""
    flag_moment = sym_elem.shape[0] > 12
    sym_elem_zero = 0*sym_elem
    sym_elem_plus_one = sym_elem_zero + 1
    sym_elem_minus_one = sym_elem_zero - 1

    flag_plus_one = sym_elem_plus_one == sym_elem
    flag_minus_one = sym_elem_minus_one == sym_elem

    str_empty = numpy.zeros(shape=sym_elem_zero.shape, dtype='<3U')

    str_empty[4] = numpy.where(flag_plus_one[4], "+"+labels[0], numpy.where(
        flag_minus_one[4], "-"+labels[0], ""))
    str_empty[5] = numpy.where(flag_plus_one[5], "+"+labels[1], numpy.where(
        flag_minus_one[5], "-"+labels[1], ""))
    str_empty[6] = numpy.where(flag_plus_one[6], "+"+labels[2], numpy.where(
        flag_minus_one[6], "-"+labels[2], ""))

    str_empty[7] = numpy.where(flag_plus_one[7], "+"+labels[0], numpy.where(
        flag_minus_one[7], "-"+labels[0], ""))
    str_empty[8] = numpy.where(flag_plus_one[8], "+"+labels[1], numpy.where(
        flag_minus_one[8], "-"+labels[1], ""))
    str_empty[9] = numpy.where(flag_plus_one[9], "+"+labels[2], numpy.where(
        flag_minus_one[9], "-"+labels[2], ""))

    str_empty[10] = numpy.where(flag_plus_one[10], "+"+labels[0], numpy.where(
        flag_minus_one[10], "-"+labels[0], ""))
    str_empty[11] = numpy.where(flag_plus_one[11], "+"+labels[1], numpy.where(
        flag_minus_one[11], "-"+labels[1], ""))
    str_empty[12] = numpy.where(flag_plus_one[12], "+"+labels[2], numpy.where(
        flag_minus_one[12], "-"+labels[2], ""))

    if flag_moment:
        str_empty[13] = numpy.where((flag_plus_one[4]*flag_plus_one[13] | 
                                     flag_minus_one[4]*flag_minus_one[13]),
                                    "+1", "-1")

    char_add = numpy.char.add
    char_strip = numpy.char.strip

    t_x = sym_elem[0]
    t_y = sym_elem[1]
    t_z = sym_elem[2]
    denom = sym_elem[3]

    gcd_x = numpy.gcd(t_x, denom)
    gcd_y = numpy.gcd(t_y, denom)
    gcd_z = numpy.gcd(t_z, denom)

    denom_x, denom_y, denom_z = denom // gcd_x, denom // gcd_y, denom // gcd_z
    t_x_new, t_y_new, t_z_new = t_x // gcd_x, t_y // gcd_y, t_z // gcd_z

    s_t_x_new, s_t_y_new = t_x_new.astype(str), t_y_new.astype(str)
    s_t_z_new = t_z_new.astype(str)

    s_denom_x, s_denom_y = denom_x.astype(str), denom_y.astype(str)
    s_denom_z = denom_z.astype(str)

    s_fract_x = char_add(char_add(s_t_x_new, "/"), s_denom_x)
    s_fract_y = char_add(char_add(s_t_y_new, "/"), s_denom_y)
    s_fract_z = char_add(char_add(s_t_z_new, "/"), s_denom_z)

    s_fract_x[numpy.char.startswith(s_fract_x, "0")] = ""
    s_fract_y[numpy.char.startswith(s_fract_y, "0")] = ""
    s_fract_z[numpy.char.startswith(s_fract_z, "0")] = ""

    flag_x = numpy.char.startswith(s_fract_x, "-")
    s_fract_x = numpy.where(flag_x, s_fract_x, char_add("+", s_fract_x))
    flag_y = numpy.char.startswith(s_fract_y, "-")
    s_fract_y = numpy.where(flag_y, s_fract_y, char_add("+", s_fract_y))
    flag_z = numpy.char.startswith(s_fract_z, "-")
    s_fract_z = numpy.where(flag_z, s_fract_z, char_add("+", s_fract_z))

    s_out_f_1 = char_add(char_add(char_add(str_empty[4], str_empty[5]),
                                  str_empty[6]), s_fract_x)
    s_out_f_2 = char_add(char_add(char_add(str_empty[7], str_empty[8]),
                                  str_empty[9]), s_fract_y)
    s_out_f_3 = char_add(char_add(char_add(str_empty[10], str_empty[11]),
                                  str_empty[12]), s_fract_z)

    s_out_f_1 = char_strip(s_out_f_1, "+")
    s_out_f_2 = char_strip(s_out_f_2, "+")
    s_out_f_3 = char_strip(s_out_f_3, "+")

    s_out_fract = char_add(char_add(char_add(char_add(s_out_f_1, ","),
                                             s_out_f_2), ","), s_out_f_3)
    if flag_moment:
        s_out = char_add(char_add(s_out_fract, ","), str_empty[13])
    else:
        s_out = s_out_fract
    return s_out


def calc_full_sym_elems(sym_elems, magn_centering):
    """Calculate full list of symmetry elements.
         0      1      2        3     4     5     6     7     8     9
    [num_x, num_y, num_z, den_xyz, r_11, r_12, r_13, r_21, r_22, r_23,
        10   11    12    13    14    15    16    17    18    19    20    21
     r_31, r_32, r_33, m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33]
    """
    new_shape = (22, ) + sym_elems.shape[1:] + magn_centering.shape[1:]
    full_sym_elems = numpy.zeros(shape=new_shape, dtype=int)
    full_ones = numpy.ones(shape=(1, ) + sym_elems.shape[1:] + magn_centering.shape[1:], dtype=int)

    n_ax = numpy.newaxis
    for j in range(2):
        full_sym_elems[4+j*9] = \
            magn_centering[4, n_ax, :] * sym_elems[4+j*9, :, n_ax] + \
            magn_centering[5, n_ax, :] * sym_elems[7+j*9, :, n_ax] + \
            magn_centering[6, n_ax, :] * sym_elems[10+j*9, :, n_ax]
        full_sym_elems[5+j*9] = \
            magn_centering[4, n_ax, :] * sym_elems[5+j*9, :, n_ax] + \
            magn_centering[5, n_ax, :] * sym_elems[8+j*9, :, n_ax] + \
            magn_centering[6, n_ax, :] * sym_elems[11+j*9, :, n_ax]
        full_sym_elems[6+j*9] = \
            magn_centering[4, n_ax, :] * sym_elems[6+j*9, :, n_ax] + \
            magn_centering[5, n_ax, :] * sym_elems[9+j*9, :, n_ax] + \
            magn_centering[6, n_ax, :] * sym_elems[12+j*9, :, n_ax]

        full_sym_elems[7+j*9] = \
            magn_centering[7, n_ax, :] * sym_elems[4+j*9, :, n_ax] + \
            magn_centering[8, n_ax, :] * sym_elems[7+j*9, :, n_ax] + \
            magn_centering[9, n_ax, :] * sym_elems[10+j*9, :, n_ax]
        full_sym_elems[8+j*9] = \
            magn_centering[7, n_ax, :] * sym_elems[5+j*9, :, n_ax] + \
            magn_centering[8, n_ax, :] * sym_elems[8+j*9, :, n_ax] + \
            magn_centering[9, n_ax, :] * sym_elems[11+j*9, :, n_ax]
        full_sym_elems[9+j*9] = \
            magn_centering[7, n_ax, :] * sym_elems[6+j*9, :, n_ax] + \
            magn_centering[8, n_ax, :] * sym_elems[9+j*9, :, n_ax] + \
            magn_centering[9, n_ax, :] * sym_elems[12+j*9, :, n_ax]

        full_sym_elems[10+j*9] = \
            magn_centering[10, n_ax, :] * sym_elems[4+j*9, :, n_ax] + \
            magn_centering[11, n_ax, :] * sym_elems[7+j*9, :, n_ax] + \
            magn_centering[12, n_ax, :] * sym_elems[10+j*9, :, n_ax]
        full_sym_elems[11+j*9] = \
            magn_centering[10, n_ax, :] * sym_elems[5+j*9, :, n_ax] + \
            magn_centering[11, n_ax, :] * sym_elems[8+j*9, :, n_ax] + \
            magn_centering[12, n_ax, :] * sym_elems[11+j*9, :, n_ax]
        full_sym_elems[12+j*9] = \
            magn_centering[10, n_ax, :] * sym_elems[6+j*9, :, n_ax] + \
            magn_centering[11, n_ax, :] * sym_elems[9+j*9, :, n_ax] + \
            magn_centering[12, n_ax, :] * sym_elems[12+j*9, :, n_ax]

    num_x, den_x = rational_sum(
        magn_centering[4, n_ax, :]*sym_elems[0, :, n_ax] +
        magn_centering[5, n_ax, :]*sym_elems[1, :, n_ax] +
        magn_centering[6, n_ax, :]*sym_elems[2, :, n_ax],
        full_ones*sym_elems[3, :, n_ax],
        full_ones*magn_centering[0, n_ax, :], full_ones*magn_centering[3, n_ax, :])

    num_y, den_y = rational_sum(
        magn_centering[7, n_ax, :]*sym_elems[0, :, n_ax] +
        magn_centering[8, n_ax, :]*sym_elems[1, :, n_ax] +
        magn_centering[9, n_ax, :]*sym_elems[2, :, n_ax],
        full_ones*sym_elems[3, :, n_ax],
        full_ones*magn_centering[1, n_ax, :], full_ones*magn_centering[3, n_ax, :])

    num_z, den_z = rational_sum(
        magn_centering[10, n_ax, :]*sym_elems[0, :, n_ax] +
        magn_centering[11, n_ax, :]*sym_elems[1, :, n_ax] +
        magn_centering[12, n_ax, :]*sym_elems[2, :, n_ax],
        full_ones*sym_elems[3, :, n_ax],
        full_ones*magn_centering[2, n_ax, :], full_ones*magn_centering[3, n_ax, :])

    num_xx = num_x*den_y*den_z
    num_yy = num_y*den_x*den_z
    num_zz = num_z*den_x*den_y
    den = den_x*den_y*den_z
    num_xx, num_yy, num_zz = num_xx % den, num_yy % den, num_zz % den

    gcd = numpy.gcd.reduce([num_xx, num_yy, num_zz, den])

    full_sym_elems[0] = num_xx // gcd
    full_sym_elems[1] = num_yy // gcd
    full_sym_elems[2] = num_zz // gcd
    full_sym_elems[3] = den // gcd

    n_vol = 1
    for val in sym_elems.shape[1:] + magn_centering.shape[1:]:
        n_vol *= val

    full_sym_elems_reshape = full_sym_elems.reshape((22, n_vol), order="C")
    full_sym_elems_unique = numpy.unique(full_sym_elems_reshape, axis=1)
    return full_sym_elems_unique


def rational_sum(numerator, denominator, *argv):
    """Sum of rational numbers."""
    if len(argv) < 2:
        gcd = numpy.gcd(numerator, denominator)
        num_out, den_out = numerator//gcd, denominator//gcd
    else:
        num_2 = argv[0]
        den_2 = argv[1]
        num_3 = numerator*den_2 + num_2*denominator
        den_3 = denominator*den_2
        gcd = numpy.gcd(num_3, den_3)
        num_out, den_out = rational_sum(num_3//gcd, den_3//gcd, argv[2:])
    return num_out, den_out


def calc_multiplicity(sym_elems, fract_xyz):
    """Calculate multiplicity for fraction xyz and given symmetry elements."""
    fract_xyz_new = calc_fract_by_sym_elem(sym_elems, fract_xyz)
    fract_xyz_round = numpy.round(fract_xyz_new, 6)
    n_xyz, n_symm, n_at = fract_xyz_round.shape
    multiplicity = numpy.array([numpy.unique(
        fract_xyz_round[:, :, i], axis=1).shape[1] for i in range(n_at)],
        dtype=int)
    return multiplicity


def calc_hkl_in_range(sym_elems, unit_cell_parameters: Tuple[float],
                      sthovl_min: float, sthovl_max: float):
    """Give index_hkl and their multiplicity in the range sthovl.

    unit_cell_parameters is (a, b, c, alpha, beta, gamma)
    """
    length_a, length_b, length_c, angle_alpha, angle_beta, angle_gamma = \
        unit_cell_parameters

    hmax = int(2.*length_a*sthovl_max)
    kmax = int(2.*length_b*sthovl_max)
    lmax = int(2.*length_c*sthovl_max)

    hmin, kmin, lmin = -1*hmax, -1*kmax, -1*lmax

    hmin = 0
    # FIXME: absent condition:
    #   shift = space_group.shift
    #   orig_x, orig_y, orig_z = zip(*shift)
    #   np_orig_x = numpy.array(orig_x, dtype=float)
    #   np_orig_y = numpy.array(orig_y, dtype=float)
    #   np_orig_z = numpy.array(orig_z, dtype=float)
    #   flag=(abs(sum(numpy.exp(2.*numpy.pi*1j*(np_orig_x*h+np_orig_y*k+ \
    #       np_orig_z*l))))>0.00001)

    r_11, r_12, r_13 = sym_elems[4], sym_elems[5], sym_elems[6]
    r_21, r_22, r_23 = sym_elems[7], sym_elems[8], sym_elems[9]
    r_31, r_32, r_33 = sym_elems[10], sym_elems[11], sym_elems[12]

    index_h, index_k, index_l = numpy.meshgrid(
        range(hmin, hmax+1, 1), range(kmin, kmax+1, 1),
        range(lmin, lmax+1, 1), indexing="ij")
    index_h, index_k = index_h.flatten(), index_k.flatten()
    index_l = index_l.flatten()

    n_a = numpy.newaxis
    index_h_new = r_11[n_a, :]*index_h[:, n_a] + \
        r_21[n_a, :]*index_k[:, n_a] + r_31[n_a, :]*index_l[:, n_a]
    index_k_new = r_12[n_a, :]*index_h[:, n_a] + \
        r_22[n_a, :]*index_k[:, n_a] + r_32[n_a, :]*index_l[:, n_a]
    index_l_new = r_13[n_a, :]*index_h[:, n_a] + \
        r_23[n_a, :]*index_k[:, n_a] + r_33[n_a, :]*index_l[:, n_a]

    keys = -10000*index_h_new + -100*index_k_new + -index_l_new

    sort_arg = numpy.argsort(keys, axis=1)
    sorted_keys = numpy.take_along_axis(keys, sort_arg, axis=1)

    unique, unique_ind = numpy.unique(sorted_keys[:, 0], return_index=True)

    flag = sorted_keys == sorted_keys[:, 0][:, n_a]
    inv_mult = flag.sum(axis=1)
    multiplicity = flag.shape[1]//inv_mult

    # ind_h_un = index_h_new[sort_arg][unique_ind]
    # ind_k_un = index_k_new[sort_arg][unique_ind]
    # ind_l_un = index_l_new[sort_arg][unique_ind]

    ind_h_un = numpy.take_along_axis(index_h_new, sort_arg, axis=1)[:, 0][
        unique_ind]
    ind_k_un = numpy.take_along_axis(index_k_new, sort_arg, axis=1)[:, 0][
        unique_ind]
    ind_l_un = numpy.take_along_axis(index_l_new, sort_arg, axis=1)[:, 0][
        unique_ind]
    mult_un = multiplicity[unique_ind]

    sthovl = calc_sthovl_by_hkl_abc_angles(
        ind_h_un, ind_k_un, ind_l_un, length_a, length_b, length_c,
        angle_alpha, angle_beta, angle_gamma)

    flag = numpy.array((sthovl >= sthovl_min) *
                       (sthovl <= sthovl_max), dtype=bool)

    sort_arg = numpy.argsort(sthovl[flag])
    ind_h_sort = ind_h_un[flag][sort_arg]
    ind_k_sort = ind_k_un[flag][sort_arg]
    ind_l_sort = ind_l_un[flag][sort_arg]
    mult_sort = mult_un[flag][sort_arg]

    return ind_h_sort, ind_k_sort, ind_l_sort, mult_sort

#         for h in range(hmin, hmax+1, 1):
#             for k in range(kmin, kmax+1, 1):
#                 for l in range(lmin, lmax+1, 1):
#                     if (flag):
#                         lhkls = [(_1, _2, _3) for _1, _2, _3 in zip(h*r_11+k*r_21+l*r_31, h*r_12+k*r_22+l*r_32, h*r_13+k*r_23+l*r_33)]
#                         lhkls.extend([(-hkl[0],-hkl[1],-hkl[2]) for hkl in lhkls])
#                         lhkls.sort(key=lambda x:10000*x[0]+100*x[1]+x[2])
#                         if (not(lhkls[-1] in lhkl)):
#                             lhkl.append(lhkls[-1])
#                             lmult.append(len(set(lhkls)))


read_magnetic_data()

# sym_elems, magn_centering = get_sym_elems_magn_centering_for_bns(71, 536)
# print("sym_elems.shape: ", sym_elems.shape)
# print("magn_centering.shape: ", magn_centering.shape)
# full_sym_elems = calc_full_sym_elems(sym_elems, magn_centering)
# print(get_str_for_sym_elem(sym_elems), end=2*"\n")
# print(get_str_for_sym_elem(magn_centering), end=2*"\n")
# print(get_str_for_sym_elem(full_sym_elems), end=2*"\n")
# fract_xyz = numpy.array([[0.1,0.0,0.0,0.2],
#                          [0.2,0.0,0.5,0.2],
#                          [0.5,0.0,0.5,0.2]], dtype=float)
# print(calc_multiplicity(full_sym_elems, fract_xyz))
# print(calc_multiplicity(full_sym_elems, (0.1, 0.2, 0.5)))
# # test calc_hkl_in_range
# half_pi = 0.5*numpy.pi
# unit_cell_parameters = (6., 6., 6., half_pi, half_pi, half_pi)
# sthovl_min, sthovl_max = 0.01, 0.70
# hkl = calc_hkl_in_range(full_sym_elems, unit_cell_parameters, sthovl_min,
#                         sthovl_max)
# for ind_h, ind_k, ind_l, mult in zip(*hkl):
#     print(f"{ind_h:4} {ind_k:4} {ind_l:4} {mult:4}")



import numpy

from scipy.special import erfc, exp1

def calc_n(alpha, beta):
    y = 0.5 * alpha * beta / (alpha + beta)
    return y

def calc_eta(h_l, h_com):
    x = h_l/h_com
    y = 1.36603 * x - 0.47719 * numpy.square(x) + 0.11116 * numpy.power(x, 3)
    return y

def calc_h_com(h_g, h_l):
    y = numpy.power(
        numpy.power(h_g, 5)
        + 2.69269 * numpy.power(h_g, 4) * h_l
        + 2.42843 * numpy.power(h_g, 3) * numpy.square(h_l)
        + 4.47163 * numpy.square(h_g) * numpy.power(h_l, 3)
        + 0.07842 * h_g * numpy.power(h_l, 4)
        + numpy.power(h_l, 5), 0.2)
    return y

def calc_h_g(sigma):
    y = numpy.sqrt(8*numpy.log(2)) * sigma
    return y

def calc_gaussian(delta_t, sigma):
    y = numpy.exp(-0.5 * numpy.square(delta_t/sigma)) / (numpy.sqrt(2 * numpy.pi) * sigma)
    return y

def calc_lorentz(delta_t, h_l):
    y = h_l / (2 * numpy.pi * (numpy.square(delta_t) + numpy.square(0.5*h_l)))
    return y

def calc_u(delta_t, sigma, alpha):
    y = 0.5 * alpha * (alpha * numpy.square(sigma) + 2*delta_t)
    return y

def calc_v(delta_t, sigma, beta):
    y = 0.5 * beta * (beta * numpy.square(sigma) - 2*delta_t)
    return y

def calc_y(delta_t, sigma, alpha):
    y = (alpha * numpy.square(sigma) + delta_t) / (sigma * numpy.sqrt(2))
    return y

def calc_z(delta_t, sigma, beta):
    y = (beta * numpy.square(sigma) - delta_t) / (sigma * numpy.sqrt(2))
    return y

def calc_e_beta_g(delta_t, sigma, alpha, beta):
    n = calc_n(alpha, beta)
    u = calc_u(delta_t, sigma, alpha)
    v = calc_v(delta_t, sigma, beta)
    y = calc_y(delta_t, sigma, alpha)
    z = calc_z(delta_t, sigma, beta)

    exp_u = numpy.exp(u)
    exp_u[numpy.isinf(exp_u)] = 1e200
    exp_v = numpy.exp(v)
    exp_v[numpy.isinf(exp_v)] = 1e200

    res = n * (exp_u * erfc(y) + exp_v * erfc(z))
    return res

def calc_p(delta_t, h_l, alpha):
    y = alpha * delta_t + 0.5* 1j * alpha * h_l
    return y

def calc_q(delta_t, h_l, beta):
    y = - beta * delta_t + 0.5* 1j * beta * h_l
    return y

def calc_e_beta_l(delta_t, h_l, alpha, beta):
    n = calc_n(alpha, beta)
    p = calc_p(delta_t, h_l, alpha)
    q = calc_q(delta_t, h_l, beta)
    exp1_p = exp1(p)
    exp1_p[numpy.isinf(exp1_p)] = 0
    exp1_p[numpy.isnan(exp1_p)] = 0
    exp_p = numpy.exp(p)
    exp_p[numpy.isinf(exp_p)] = 1e200
    exp1_q = exp1(q)
    exp1_q[numpy.isinf(exp1_q)] = 0
    exp1_q[numpy.isnan(exp1_q)] = 0
    exp_q = numpy.exp(q)
    exp_q[numpy.isinf(exp_q)] = 1e200
    res = -2 * n / numpy.pi * (
        numpy.imag(exp_p * exp1_p) +
        numpy.imag(exp_q * exp1_q)
    )
    return res

def calc_profile(delta_t, sigma, h_l, alpha, beta_0, beta_1, r_0):
    h_g = calc_h_g(sigma)
    h_com = calc_h_com(h_g, h_l)
    eta = calc_eta(h_l, h_com)
    e_beta_0_g = calc_e_beta_g(delta_t, sigma, alpha, beta_0)
    e_beta_1_g = calc_e_beta_g(delta_t, sigma, alpha, beta_1)
    e_beta_0_l = calc_e_beta_l(delta_t, h_l, alpha, beta_0)
    e_beta_1_l = calc_e_beta_l(delta_t, h_l, alpha, beta_1)
    pv_0 = eta * e_beta_0_l + (1. - eta) * e_beta_0_g
    pv_1 = eta * e_beta_1_l + (1. - eta) * e_beta_1_g
    y = r_0 * pv_0 + (1.-r_0) * pv_1
    return y


# Z-code models

def calc_sigma_square(d, sigma_0_sq, sigma_1_sq, sigma_2_sq):
    y = sigma_0_sq + sigma_1_sq * numpy.square(d) + sigma_2_sq * numpy.power(d, 4)
    return y


def calc_h_l(d, gamma_0, gamma_1, gamma_2):
    y = gamma_0 + gamma_1 * d + gamma_2 * numpy.square(d)
    return y


def calc_r_0(d, r_01, r_02, r_03):
    """Switching function
    """
    y = r_01 * numpy.exp(-r_02 * numpy.power(d, -r_03))
    return y


def calc_alpha_prime(d, alpha_1, alpha_2):
    y = 1./(alpha_1 + alpha_2/d)
    return y

def calc_beta_0_prime(d, beta_00, beta_01):
    y = beta_00 + beta_01/d
    return y


def calc_beta_1_prime(d, beta_10):
    y = beta_10
    return y


def calc_profile_by_zcode_parameters(
        delta_t, d,
        sigma_0_sq, sigma_1_sq, sigma_2_sq,
        gamma_0, gamma_1, gamma_2,
        r_01, r_02, r_03,
        alpha_1, alpha_2,
        beta_00, beta_01,
        beta_10):
    r_0 = calc_r_0(d, r_01, r_02, r_03)
    alpha_prime = calc_alpha_prime(d, alpha_1, alpha_2)
    beta_0_prime = calc_beta_0_prime(d, beta_00, beta_01)
    beta_1_prime = calc_beta_1_prime(d, beta_10)
    sigma = numpy.sqrt(calc_sigma_square(d, sigma_0_sq, sigma_1_sq, sigma_2_sq))
    h_l = calc_h_l(d, gamma_0, gamma_1, gamma_2)

    h_g = calc_h_g(sigma)
    h_com = calc_h_com(h_g, h_l)
    alpha = 1./(alpha_prime * h_com)
    beta_0 = 1./(beta_0_prime * h_com)
    beta_1 = 1./(beta_1_prime * h_com)
    """
    alpha = alpha_prime
    beta_0 = beta_0_prime
    beta_1 = beta_1_prime
    """
    res = calc_profile(delta_t, sigma, h_l, alpha, beta_0, beta_1, r_0)
    return res


#!/usr/bin/env python
"""
CrysPy is a crystallographic library for neutron data analysis. It allows to
refine polarized neutron diffraction experiments performed with single crystals
as well as with powder magnetic compounds. The program has been mainly
developed for Rietveld analysis of polarized neutron powder diffraction data.
The program can be also used as a MEM tool to reconstruct magnetization
(or spin) density. Time-of-flight (TOF) neutron data analysis is also
available.

List of functions is accessible by command:
    >>> cryspy.functions()

List of classes is accessible by command:
    >>> cryspy.classes()

To search the specific function or the class key words should be used:
    >>> cryspy.functions("flip ratio")
    >>> cryspy.classes("tof")

To load data from ".cif" file use the function:
    >>> data = cryspy.load_file(file_name)

where `file_name` is the address of the file.
    
    
Features
  - Analysis of the polarized neutron scattering on crystals;
  - Diffraction data refinement for single crystals or powder by RhoChi class;
  - MEM for reconstruction of magnetization density by MEM class;
  - Powder refinement of magnetic space groups.
  - TOF neutron data analysis

For more information please see the site: https://www.cryspy.fr/
"""
__author__ = 'Iurii KIBALIN'
__copyright__   = "Copyright 2024, "
__credits__ = ["Iurii KIBALIN", "Andrew SAZONOV", "Arsen GOUKASSOV"]
__license__ = "GPL"
__version__ = "0.7.7"
__maintainer__ = "Iurii KIBALIN"
__email__ = "iurii.kibalin@cea.fr"
__status__ = "Development"
__date__ = "12.09.2024"
name = "cryspy"


from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.cl_4_global import GlobalN


from cryspy.A_functions_base.function_1_algebra import \
    calc_scalar_product_by_vectors,\
    calc_scalar_product_by_complex_vectors,\
    calc_modulus_sq_by_complex_vector

from cryspy.A_functions_base.function_1_atomic_vibrations import calc_beta_by_u, vibration_constraints, apply_constraint_on_cell_by_type_cell
from cryspy.A_functions_base.function_1_error_simplex import error_estimation_simplex
from cryspy.A_functions_base.function_1_gamma_nu import gammanu_to_tthphi, tthphi_to_gammanu, recal_int_to_tthphi_grid, recal_int_to_gammanu_grid

from cryspy.A_functions_base.function_1_inversed_hessian import estimate_inversed_hessian_matrix
from cryspy.A_functions_base.function_1_magnetic import get_j0_j2_by_symbol
from cryspy.A_functions_base.function_1_markdown import md_to_html

from cryspy.A_functions_base.function_1_matrices import \
    calc_chi_sq, tri_linear_interpolation, transform_string_to_r_b, transform_string_to_digits,\
    transform_fraction_with_label_to_string, transform_digits_to_string, transform_r_b_to_string, \
    calc_mRmCmRT, calc_rotation_matrix_ij_by_euler_angles,\
    calc_euler_angles_by_rotation_matrix_ij, calc_determinant_matrix_ij, \
    calc_inverse_matrix_ij, calc_rotation_matrix_ij_around_axis, \
    calc_product_matrices, calc_product_matrix_vector, calc_vector_angle,\
    calc_vector_product, scalar_product, calc_rotation_matrix_by_two_vectors, \
    ortogonalize_matrix, calc_moment_2d_by_susceptibility, calc_phase_3d
from cryspy.A_functions_base.function_1_objects import get_functions_of_objet, \
    variable_name_to_string, change_variable_name, get_table_html_for_variables
from cryspy.A_functions_base.function_1_roots import calc_roots
from cryspy.A_functions_base.function_1_scat_length_neutron import \
    apply_constraint_on_cell_by_type_cell, get_scat_length_neutron

from cryspy.A_functions_base.function_1_strings import value_error_mark_to_string, \
    ciftext_to_html, find_prefix, common_string, string_to_value_error_mark, \
    transform_string_to_r_b, transform_string_to_digits,\
    transform_fraction_with_label_to_string, transform_digits_to_string,\
    transform_r_b_to_string

from cryspy.A_functions_base.powder_diffraction_tof import tof_Jorgensen, \
    tof_Jorgensen_VonDreele


from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_volume_uc_by_abc_cosines,\
    calc_volume_uc_by_abc_angles,\
    calc_inverse_d_by_hkl_abc_cosines,\
    calc_inverse_d_by_hkl_abc_angles,\
    calc_sthovl_by_hkl_abc_cosines,\
    calc_sthovl_by_hkl_abc_angles,\
    calc_phase_3d,\
    calc_moment_2d_by_susceptibility,\
    ortogonalize_matrix

from cryspy.A_functions_base.function_2_mem import \
    calc_asymmetric_unit_cell_indexes, calc_index_atom_symmetry_closest_to_fract_xyz, \
    calc_factor_in_front_of_density_for_fm, calc_moment_perp,\
    transfer_to_density_3d, transfer_to_chi_3d

from cryspy.A_functions_base.function_2_space_group import \
    get_crystal_system_by_it_number, get_name_hm_short_by_it_number, \
    get_name_schoenflies_by_it_number, get_name_hall_by_it_number, \
    get_name_hm_extended_by_it_number_it_coordinate_system_code, \
    get_name_hm_full_by_it_number

from cryspy.A_functions_base.function_2_sym_elems import \
    form_symm_elems_by_b_i_r_ij, transform_to_p1, get_string_by_symm_elem

from cryspy.A_functions_base.function_3_den_file import \
    read_den_file, save_to_den_file

from cryspy.A_functions_base.function_3_extinction import \
    calc_extinction,\
    calc_extinction_2

from cryspy.A_functions_base.function_3_mcif import \
    calc_fract_by_sym_elem, calc_moment_by_sym_elem

from cryspy.A_functions_base.function_4_flip_ratio import \
    calc_f_plus_sq, calc_f_minus_sq, calc_flip_ratio

from cryspy.B_parent_classes.cl_2_loop import \
 get_prefix_of_loop

from cryspy.D_functions_item_loop.function_1_section_from_density_point \
    import calc_section_from_density_point


from cryspy.D_functions_item_loop.function_1_report_magnetization_ellipsoid import \
    magnetization_ellipsoid_by_u_ij, report_main_axes_of_magnetization_ellipsoids

# from cryspy.D_functions_item_loop.structure_factor import \
#     calculate_nuclear_structure_factor, calculate_structure_factor_tensor

from cryspy.H_functions_global.function_1_cryspy_objects import \
    str_to_globaln, file_to_globaln, str_to_items

from cryspy.H_functions_global.function_1_cryspy_objects import \
    file_to_globaln as load_file

from cryspy.H_functions_global.function_1_cryspy_objects import \
    L_GLOBAL_CLASS, L_DATA_CLASS, L_LOOP_CLASS, L_ITEM_CLASS, load_packages, \
    add_package, packages, delete_package, L_FUNCTION_ADD

from cryspy.H_functions_global.powder_experiments import \
    report_powder_experiments

from cryspy.procedure_rhochi.rhochi import rhochi_rietveld_refinement, \
    rhochi_rietveld_refinement_with_parameters, \
    rhochi_no_refinement, rhochi_inversed_hessian


from cryspy.procedure_mempy.mempy import  mempy_magnetization_density_reconstruction, \
    mempy_spin_density_reconstruction,\
    mempy_reconstruction_with_parameters,\
    mempy_cycle_with_parameters




def repr_function(function, flag_long: bool = False):
    ls_out = []
    name = function.__name__
    code = function.__code__
    defaults_val = function.__defaults__
    if defaults_val is None:
        len_defaults_val = 0
    else:
        len_defaults_val = len(defaults_val)
    argcount = code.co_argcount
    varnames_in = code.co_varnames[:argcount]
    if len_defaults_val == 0:
        s_in_1 = ", ".join([name for name in varnames_in])
    else:
        s_in_1 = ", ".join([name for name in varnames_in[:-len_defaults_val]])
        
    if len_defaults_val == 0:
        s_in_2 = ""
    else:
        s_in_2 = ", ".join([f"{name:}=..." for name in varnames_in[-len_defaults_val:]])

    if ((s_in_1 == "") & (s_in_2 == "")):
        ls_out.append(f"Function: {name:}()")
    elif ((s_in_1 == "") & (s_in_2 != "")):
        ls_out.append(f"Function: {name:}({s_in_2:})")
    elif ((s_in_1 != "") & (s_in_2 == "")):
        ls_out.append(f"Function: {name:}({s_in_1:})")
    else:
        ls_out.append(f"Function: {name:}({s_in_1:}, {s_in_2:})")
    doc = function.__doc__
    if doc is not None:
        if flag_long:
            ls_out.exted([f"    {line:}" for line in doc.split("\n")])
        else:
            
            ls_out.append("    "+doc.strip().split("\n")[0])
    return "\n".join(ls_out)


L_FUNCTION = [
    calc_scalar_product_by_vectors,
    calc_scalar_product_by_complex_vectors,
    calc_modulus_sq_by_complex_vector,
    calc_beta_by_u, vibration_constraints, apply_constraint_on_cell_by_type_cell,
    error_estimation_simplex,
    gammanu_to_tthphi, tthphi_to_gammanu, recal_int_to_tthphi_grid, recal_int_to_gammanu_grid,
    estimate_inversed_hessian_matrix,
    get_j0_j2_by_symbol,
    md_to_html,
    calc_chi_sq, tri_linear_interpolation, transform_string_to_r_b, transform_string_to_digits,
    transform_fraction_with_label_to_string, transform_digits_to_string, transform_r_b_to_string,
    calc_mRmCmRT, calc_rotation_matrix_ij_by_euler_angles,
    calc_euler_angles_by_rotation_matrix_ij, calc_determinant_matrix_ij,
    calc_inverse_matrix_ij, calc_rotation_matrix_ij_around_axis,
    calc_product_matrices, calc_product_matrix_vector, calc_vector_angle,
    calc_vector_product, scalar_product, calc_rotation_matrix_by_two_vectors,
    ortogonalize_matrix, calc_moment_2d_by_susceptibility, calc_phase_3d,
    get_functions_of_objet,
    variable_name_to_string, change_variable_name, get_table_html_for_variables,
    calc_roots,
    apply_constraint_on_cell_by_type_cell, get_scat_length_neutron,
    value_error_mark_to_string, 
    ciftext_to_html, find_prefix, common_string, string_to_value_error_mark,
    transform_string_to_r_b, transform_string_to_digits,
    transform_fraction_with_label_to_string, transform_digits_to_string,
    transform_r_b_to_string,
    tof_Jorgensen,
    tof_Jorgensen_VonDreele,
    calc_volume_uc_by_abc_cosines,
    calc_volume_uc_by_abc_angles,
    calc_inverse_d_by_hkl_abc_cosines,
    calc_inverse_d_by_hkl_abc_angles,
    calc_sthovl_by_hkl_abc_cosines,
    calc_sthovl_by_hkl_abc_angles,
    calc_phase_3d,
    calc_moment_2d_by_susceptibility,
    ortogonalize_matrix,
    calc_asymmetric_unit_cell_indexes, calc_index_atom_symmetry_closest_to_fract_xyz, 
    calc_factor_in_front_of_density_for_fm, calc_moment_perp,
    transfer_to_density_3d, transfer_to_chi_3d,
    get_crystal_system_by_it_number, get_name_hm_short_by_it_number, 
    get_name_schoenflies_by_it_number, get_name_hall_by_it_number, 
    get_name_hm_extended_by_it_number_it_coordinate_system_code, 
    get_name_hm_full_by_it_number,
    form_symm_elems_by_b_i_r_ij, transform_to_p1, get_string_by_symm_elem,
    read_den_file, save_to_den_file,
    calc_extinction,
    calc_extinction_2,
    calc_fract_by_sym_elem, calc_moment_by_sym_elem,
    calc_f_plus_sq, calc_f_minus_sq, calc_flip_ratio,
    calc_section_from_density_point,
    str_to_globaln, file_to_globaln, str_to_items,
    load_file,
    load_packages,
    add_package, packages, delete_package,
    repr_function,
    magnetization_ellipsoid_by_u_ij,
    report_main_axes_of_magnetization_ellipsoids,
    report_powder_experiments,
    rhochi_rietveld_refinement, 
    rhochi_rietveld_refinement_with_parameters, 
    rhochi_no_refinement,
    rhochi_inversed_hessian,
    mempy_magnetization_density_reconstruction, 
    mempy_spin_density_reconstruction,
    mempy_reconstruction_with_parameters,
    mempy_cycle_with_parameters,
    get_prefix_of_loop]

def functions(s_name: str = "", flag_long: bool = False):
    l_function = L_FUNCTION + L_FUNCTION_ADD
    ls_out = []
    for function in l_function:
        cond_1 = function.__name__.lower().find(s_name.lower()) != -1
        if function.__doc__ is None:
            cond_2 = False
        else:
            cond_2 = function.__doc__.lower().find(s_name.lower()) != -1
        if cond_1:
            ls_out.append(repr_function(function, flag_long=flag_long))
        elif cond_2:
            ls_out.append(repr_function(function, flag_long=flag_long))
    s_out = "\n\n".join(ls_out)
    print(s_out)

def repr_class(obj_class):
    s_name = obj_class.__name__
    doc = obj_class.__doc__
    if doc is None:
        s_doc = ""
    else:
        s_doc = doc.strip().split("\n")[0]
    return f"|{s_name:30}|{s_doc:70}|"
    
def classes(s_name: str = ""):
    l_class = L_GLOBAL_CLASS+L_DATA_CLASS+L_LOOP_CLASS+L_ITEM_CLASS
    ls_out = ["|class                         |description                                                           |"]
    ls_out.append("|------------------------------|----------------------------------------------------------------------|")
    for class_ in l_class:
        if class_.__name__.lower().find(s_name.lower()) != -1:
            ls_out.append(repr_class(class_))
    s_out = "\n".join(ls_out)
    print(s_out)


def __getattr__(name):
    for obj_class in L_ITEM_CLASS + L_LOOP_CLASS + L_DATA_CLASS + \
            L_GLOBAL_CLASS + L_FUNCTION_ADD:
        if name == obj_class.__name__:
            return obj_class
    raise AttributeError(f"module 'cryspy' has no attribute '{name:}'")

    


"""Functions for magnetic section.

Functions
---------
    - get_j0_j2_by_symbol
"""
import numpy

from cryspy.A_functions_base.database import DATABASE
from cryspy.A_functions_base.charge_form_factor import get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol


def get_j0_j2_parameters(symbols):
    """Get <j0>, <j2> parameters by symbols."""
    j0_parameters = numpy.zeros((7, symbols.shape[0]), dtype=float)
    j2_parameters = numpy.zeros((7, symbols.shape[0]), dtype=float)
    d_mff = DATABASE["Magnetic Form Factor, tabulated"]
    for i_symbol, symbol in enumerate(symbols):
        atom_name, ion_charge, isotope_number = get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(symbol)
        try:
            j0_parameters[:, i_symbol] = d_mff[("j0_AaBbCcD", atom_name, ion_charge)]
        except KeyError:
            pass
        try:
            j2_parameters[:, i_symbol] = d_mff[("j2_AaBbCcD", atom_name, ion_charge)]
        except KeyError:
            pass
    return j0_parameters, j2_parameters

def calc_j0(sthovl, kappa, j0_parameters,
        flag_sthovl: bool=False, flag_kappa: bool=False):
    """Calculate j0."""
    j0_A = j0_parameters[0]
    j0_a = j0_parameters[1]
    j0_B = j0_parameters[2]
    j0_b = j0_parameters[3]
    j0_C = j0_parameters[4]
    j0_c = j0_parameters[5]
    j0_D = j0_parameters[6]
    _h = numpy.square(sthovl/kappa)
    j0_av = (j0_A*numpy.exp(-j0_a*_h) +
             j0_B*numpy.exp(-j0_b*_h) +
             j0_C*numpy.exp(-j0_c*_h)+j0_D)
    dder = {}
    if flag_sthovl:
        s_k = sthovl/kappa
        dder["sthovl"] = -2.*s_k*(j0_A*numpy.exp(-j0_a*_h)*j0_a +
             j0_B*numpy.exp(-j0_b*_h)*j0_b +
             j0_C*numpy.exp(-j0_c*_h)*j0_c)/kappa
    if flag_kappa:
        s_k = sthovl/kappa
        dder["kappa"] = 2.*_h*(j0_A*numpy.exp(-j0_a*_h)*j0_a +
             j0_B*numpy.exp(-j0_b*_h)*j0_b +
             j0_C*numpy.exp(-j0_c*_h)*j0_c)/kappa
    return j0_av, dder

def calc_j2(sthovl, kappa, j2_parameters,
        flag_sthovl: bool=False, flag_kappa: bool=False):
    """Calculate j2."""
    j2_A = j2_parameters[0]
    j2_a = j2_parameters[1]
    j2_B = j2_parameters[2]
    j2_b = j2_parameters[3]
    j2_C = j2_parameters[4]
    j2_c = j2_parameters[5]
    j2_D = j2_parameters[6]
    _h = numpy.square(sthovl/kappa)
    j2_av = (j2_A*numpy.exp(-j2_a*_h) +
             j2_B*numpy.exp(-j2_b*_h) +
             j2_C*numpy.exp(-j2_c*_h) + j2_D)*_h
    dder = {}
    if flag_sthovl:
        s_k = sthovl/kappa
        dder["sthovl"] = -2.*s_k*_h*(j2_A*numpy.exp(-j2_a*_h)*j2_a +
             j2_B*numpy.exp(-j2_b*_h)*j2_b +
             j2_C*numpy.exp(-j2_c*_h)*j2_c)/kappa + (j2_A*numpy.exp(-j2_a*_h) +
             j2_B*numpy.exp(-j2_b*_h) +
             j2_C*numpy.exp(-j2_c*_h) + j2_D)/kappa
    if flag_kappa:
        s_k = sthovl/kappa
        dder["kappa"] = 2.*_h*_h*(j2_A*numpy.exp(-j2_a*_h)*j2_a +
             j2_B*numpy.exp(-j2_b*_h)*j2_b +
             j2_C*numpy.exp(-j2_c*_h)*j2_c)/kappa - s_k*(j2_A*numpy.exp(-j2_a*_h) +
             j2_B*numpy.exp(-j2_b*_h) +
             j2_C*numpy.exp(-j2_c*_h) + j2_D)/kappa
    return j2_av, dder


def calc_form_factor(
        sthovl, lande_factor, kappa, j0_parameters, j2_parameters,
        flag_only_orbital=False,
        flag_sthovl: bool=False, flag_kappa: bool=False,
        flag_lande_factor: bool=False):
    r"""Calculate magnetic form factor in frame of Spherical model.
    (Int.Tabl.C.p.592)

    Calculation of magnetic form factor
    `<https://journals.aps.org/prb/pdf/10.1103/PhysRevB.79.140405>`
    mismatch with international tables where (1.0-2.0/np_factor_lande)
    """
    # not sure about kappa, it is here just for test, by default it is 1.0
    j2_av, dder_j2 = calc_j2(
        sthovl, kappa, j2_parameters, flag_sthovl=flag_sthovl,
        flag_kappa=flag_kappa)
    form_factor_orbital = (2.0/lande_factor-1.0)*j2_av

    if flag_only_orbital:
        form_factor = form_factor_orbital
    else:
        j0_av, dder_j0 = calc_j0(
            sthovl, kappa, j0_parameters, flag_sthovl=flag_sthovl,
            flag_kappa=flag_kappa)
        form_factor = j0_av + form_factor_orbital
    dder = {}
    if flag_sthovl:
        if flag_only_orbital:
            dder["sthovl"] = dder_j2["sthovl"]*(2.0/lande_factor-1.0)
        else:
            dder["sthovl"] = dder_j0["sthovl"] + dder_j2["sthovl"]*(2.0/lande_factor-1.0)
    if flag_kappa:
        if flag_only_orbital:
            dder["kappa"] = dder_j2["kappa"]*(2.0/lande_factor-1.0)
        else:
            dder["kappa"] = dder_j0["kappa"] + dder_j2["kappa"]*(2.0/lande_factor-1.0)
    if flag_lande_factor:
        dder["lande_factor"]=-2.0*j2_av/numpy.square(lande_factor)
    return form_factor, dder


"""Pd2dBackground class."""
import numpy
import scipy
import scipy.interpolate
from typing import NoReturn, Union

from cryspy.A_functions_base.function_1_strings import \
    string_to_value_error_mark, value_error_mark_to_string

from cryspy.B_parent_classes.cl_1_item import ItemN

from cryspy.C_item_loop_classes.cl_1_pd2d_meas import Pd2dMeas

na = numpy.newaxis

class Pd2dBackground(ItemN):
    """
    Pd2dBackground class.

    PdInstrReflexAsymmetry describes asymmetry of Bragg reflections for
    1d powder diffractometer.

    Attributes
    ----------
        - ttheta_phi_intensity

    Internal Attributes
    -------------------
        - ttheta, phi, intensity, intensity_sigma
        - intensity_refinement, intensity_constraint

    Methods
    -------
        - get_variable_names
        - get_variable_by_name, set_variable_by_name
        - form_ttheta_phi_intensity
        - interpolate_by_points
    """

    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    # ("matrix", "matrix", "matrix", "matrix")
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = ("gamma_nu_intensity", )
    ATTR_OPTIONAL_TYPES = (str, str)
    ATTR_OPTIONAL_CIF = ("gamma_nu_intensity", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("intensity", "intensity_sigma",
                      "intensity_refinement", "intensity_constraint",
                      "gamma", "nu")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd2d_background"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dBackground, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self):
        """Form object."""
        self_keys = self.__dict__.keys()

        l_1 = (self.gamma_nu_intensity).strip().split("\n")
        l_gamma = [float(_) for _ in l_1[0].strip().split()[1:]]
        l_nu, ll_intensity = [], []
        for line in l_1[1:]:
            l_1 = line.strip().split()
            l_nu.append(float(l_1[0]))
            ll_intensity.append(l_1[1:])
        ll_intensity_2 = [[string_to_value_error_mark(ll_intensity[_2][_1])[:2]
                         for _2 in range(len(ll_intensity))]
                        for _1 in range(len(ll_intensity[0]))]
        np_int_sigma = numpy.array(ll_intensity_2, dtype=float)
        ll_mark = [[string_to_value_error_mark(ll_intensity[_2][_1])[2]
                         for _2 in range(len(ll_intensity))]
                        for _1 in range(len(ll_intensity[0]))]
        np_int_mark = numpy.array(ll_mark, dtype=str)
        self.__dict__["gamma"] = l_gamma
        self.__dict__["nu"] = l_nu
        self.__dict__["intensity"] = np_int_sigma[:, :, 0]
        self.__dict__["intensity_sigma"] = numpy.where(
            numpy.isnan(np_int_sigma[:, :, 1]), 0., np_int_sigma[:, :, 1])
        self.__dict__["intensity_refinement"] = numpy.where(
            numpy.isnan(np_int_sigma[:, :, 1]), False, True)
        self.__dict__["intensity_constraint"] = numpy.zeros(
            shape=np_int_sigma[:, :, 0].shape, dtype=bool)
        self.__dict__["intensity_mark"] = np_int_mark

    def form_gamma_nu_intensity(self) -> NoReturn:
        """Form 2theta_phi_intensity from internal attributes."""
        if ((self.nu is not None) & (self.gamma is not None) &
                (self.intensity is not None)):
            ls_out = []
            ls_out.append(f"{len(self.nu):12} " + " ".join(
                [f"{_:6.2f}      " for _ in self.gamma]))
            ll_intensity = self.intensity
            ll_intensity_sigma = self.intensity_sigma
            ll_intensity_refinement = self.intensity_refinement
            ll_intensity_mark = self.intensity_mark

            ll_intensity = [[ll_intensity[_2][_1] for _2 in
                             range(len(ll_intensity))] for _1 in
                            range(len(ll_intensity[0]))]
            ll_intensity_sigma = [[ll_intensity_sigma[_2][_1] for _2 in
                                   range(len(ll_intensity_sigma))] for _1 in
                                  range(len(ll_intensity_sigma[0]))]
            ll_intensity_refinement = [[ll_intensity_refinement[_2][_1] for _2
                                        in range(len(ll_intensity_refinement))]
                                       for _1 in range(len(
                                               ll_intensity_refinement[0]))]
            ll_intensity_mark = [[ll_intensity_mark[_2][_1] for _2
                                        in range(len(ll_intensity_mark))]
                                       for _1 in range(len(
                                               ll_intensity_mark[0]))]
            for nu, l_int, l_int_sig, l_int_ref, l_int_mark in \
                zip(self.nu, ll_intensity, ll_intensity_sigma,
                    ll_intensity_refinement, ll_intensity_mark):
                ls_out.append("{:12.2f} ".format(nu) +
                              " ".join(
                                  [f"{value_error_mark_to_string(int, int_sig, int_mark):12}"
                                   if int_ref else f"{int:12}"
                                   for int, int_sig, int_ref, int_mark in
                                   zip(l_int, l_int_sig, l_int_ref, l_int_mark)]))

            self.__dict__["gamma_nu_intensity"] = "\n".join(ls_out)

    def get_variable_names(self) -> list:
        """
        Get names of variable as a list.

        (((#prefix, #NAME), (#attribute, (#index_1, #index_2)))

        Returns
        -------
        list
            List of names of variable.

        """
        prefix = self.PREFIX
        if self.is_attribute("intensity"):
            np_i, np_j = numpy.where(self.intensity_refinement == True)
            return [((prefix, None), ("intensity", (i, j)))
                    for i, j in zip(np_i, np_j)]
        else:
            return []

    def get_variable_by_name(self, name: tuple) -> Union[float, int, str]:
        """
        Get variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, ), (#attribute, (#index_1, #index_2)))

        Returns
        -------
        Union[float, int, str]
            Value.

        """
        prefix_t = name[0]
        if prefix_t[0] != self.PREFIX:
            return None
        if len(name) == 1:
            return self
        attr_t = name[1]
        attr_name, ind_ij = attr_t
        return getattr(self, attr_name)[ind_ij]

    def set_variable_by_name(self, name: tuple, value) -> NoReturn:
        """
        Set value to variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, ), (#attribute, ))
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn

        """
        prefix_t, attr_t = name
        if prefix_t[0] != self.PREFIX:
            return
        attr_name, ind_ij = attr_t
        np_val = getattr(self, attr_name)
        np_val[ind_ij] = value

    def interpolate_by_points(self, tth, phi):
        """Interpolate by points."""
        l_phi_b = self.nu
        l_tth_b = self.gamma
        ll_int_b = self.intensity
        ll_int_b = [[float(ll_int_b[_2][_1]) for _2 in range(len(ll_int_b))]
                    for _1 in range(len(ll_int_b[0]))]
        if len(l_tth_b) == 0:
            int_2d = numpy.zeros((tth.size, phi.size), dtype=float)
        else:
            phi_b = numpy.array(l_phi_b, dtype=float)
            tth_b = numpy.array(l_tth_b, dtype=float)
            int_b = numpy.array(ll_int_b, dtype=float)
            func = scipy.interpolate.interp2d(tth_b, phi_b, int_b)
            # tth_2d, phi_2d = numpy.meshgrid(tth, phi, indexing="ij")
            int_2d = func(tth, phi)
            int_2d = int_2d.transpose()
        return int_2d

    def is_variables(self):
        """
        Redefine function
        """
        return numpy.any(self.intensity_refinement)


    def define_points(self, pd2d_meas: Pd2dMeas, step_gamma: float = 10., step_nu: float = 10.):
        gamma = pd2d_meas.gamma
        nu = pd2d_meas.nu
        if pd2d_meas.is_attribute("gamma_nu_intensity_plus"):
            intensity = (pd2d_meas.intensity_plus + pd2d_meas.intensity_minus)
        else:
            intensity = pd2d_meas.intensity

        points_tth = int((gamma.max()-gamma.min())/step_gamma + 2)
        points_phi = int((nu.max()-nu.min())/step_nu + 2)

        ttheta_bkgr = numpy.linspace(gamma.min(), gamma.max(), points_tth, endpoint=True)
        phi_bkgr = numpy.linspace(nu.min(), nu.max(), points_phi, endpoint=True)

        flags_tth = numpy.abs(gamma[na, :]-ttheta_bkgr[:, na])<step_gamma
        flags_phi = numpy.abs(nu[na, :]-phi_bkgr[:, na])<step_nu
        flags = flags_tth[:, :, na, na] * flags_phi[na, na, :, :]
        int_bkrg = numpy.zeros((points_tth, points_phi), dtype=float)
        for i_tth in range(points_tth):
            for i_phi in range(points_phi):
                flag = flags[i_tth, :, i_phi, :]
                flag_nan = numpy.isnan(intensity[flag])
                if numpy.all(flag_nan):
                    int_bkrg[i_tth, i_phi] = 0.
                else:
                    int_bkrg[i_tth, i_phi] = numpy.nanmin(intensity[flag][numpy.logical_not(flag_nan)])
        self.gamma = ttheta_bkgr
        self.nu = phi_bkgr
        self.intensity = numpy.round(int_bkrg, decimals=5)
        self.intensity_refinement = numpy.zeros(int_bkrg.shape, dtype=bool)
        self.intensity_sigma = numpy.zeros(int_bkrg.shape, dtype=float)
        self.intensity_mark = numpy.zeros(int_bkrg.shape, dtype=str)
        self.form_gamma_nu_intensity()
        return

    def get_dictionary(self):
        res = {}
        res["background_gamma"] = numpy.array(self.gamma, dtype=float)*numpy.pi/180.
        res["background_nu"] = numpy.array(self.nu, dtype=float)*numpy.pi/180.
        res["background_intensity"] = numpy.array(self.intensity, dtype=float)
        res["flags_background_intensity"] = numpy.array(self.intensity_refinement, dtype=bool)
        return res


"""Description of classes Section, SectionL."""
from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_atoms_in_unit_cell
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_space_group_symop import \
    SpaceGroupSymop, SpaceGroupSymopL


class Section(ItemN):
    """Section description.

    Describe information concerning the density point.
    """

    ATTR_MANDATORY_NAMES = ("id", "size_x", "size_y", "atom_center",
                            "atom_along_axis_x", "atom_x_operation_xyz",
                            "atom_along_axis_y", "atom_y_operation_xyz",
                            "points_x", "points_y", "url_out")
    ATTR_MANDATORY_TYPES = (str, float, float, str, str, str, str, str, int,
                            int, str)
    ATTR_MANDATORY_CIF = ("id", "size_x", "size_y", "atom_center",
                          "atom_along_axis_x", "atom_x_operation_xyz",
                          "atom_along_axis_y", "atom_y_operation_xyz",
                          "points_x", "points_y", "url_out")

    ATTR_OPTIONAL_NAMES = ("field_x", "field_y", "field_z")
    ATTR_OPTIONAL_TYPES = (float, float, float)
    ATTR_OPTIONAL_CIF = ("field_x", "field_y", "field_z")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"size_x": "{:.2f}", "size_y": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "section"

    def __init__(self, **kwargs) -> NoReturn:
        super(Section, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_axes_x_y_z(self, cell: Cell, atom_site: AtomSiteL):
        """
        Calculate three vectors of axes: pos_x, pos_y, pos_z.

        Arguments
        ---------
            - cell
            - atom_site (loop)
        """
        atom_o_label = self.atom_center
        _item_a_s_o = atom_site[atom_o_label]
        fract_a_o_xyz = numpy.array([_item_a_s_o.fract_x, _item_a_s_o.fract_y,
                                     _item_a_s_o.fract_z], dtype=float)
        pos_a_o_xyz = cell.calc_position_by_coordinate(
            fract_a_o_xyz[0], fract_a_o_xyz[1], fract_a_o_xyz[2])
        pos_a_o_xyz = numpy.array(pos_a_o_xyz, dtype=float)

        atom_x_label = self.atom_along_axis_x
        atom_x_operation_xyz = self.atom_x_operation_xyz
        s_g_s_a_x = SpaceGroupSymop(operation_xyz=atom_x_operation_xyz)
        _item_a_s_x = atom_site[atom_x_label]
        fract_a_x_xyz = numpy.array([_item_a_s_x.fract_x, _item_a_s_x.fract_y,
                                     _item_a_s_x.fract_z], dtype=float)
        fract_a_x_xyz = numpy.dot(s_g_s_a_x.r, fract_a_x_xyz) + s_g_s_a_x.b
        pos_a_x_xyz = cell.calc_position_by_coordinate(
            fract_a_x_xyz[0], fract_a_x_xyz[1], fract_a_x_xyz[2])
        pos_a_x_xyz = numpy.array(pos_a_x_xyz, dtype=float)

        atom_y_label = self.atom_along_axis_y
        atom_y_operation_xyz = self.atom_y_operation_xyz
        s_g_s_a_y = SpaceGroupSymop(operation_xyz=atom_y_operation_xyz)
        _item_a_s_y = atom_site[atom_y_label]
        fract_a_y_xyz = numpy.array([_item_a_s_y.fract_x, _item_a_s_y.fract_y,
                                     _item_a_s_y.fract_z], dtype=float)
        fract_a_y_xyz = numpy.dot(s_g_s_a_y.r, fract_a_y_xyz) + s_g_s_a_y.b
        pos_a_y_xyz = cell.calc_position_by_coordinate(
            fract_a_y_xyz[0], fract_a_y_xyz[1], fract_a_y_xyz[2])
        pos_a_y_xyz = numpy.array(pos_a_y_xyz, dtype=float)

        v_pos_x = (pos_a_x_xyz-pos_a_o_xyz)/((
            numpy.square(pos_a_x_xyz-pos_a_o_xyz)).sum())**0.5
        v_pos_a_y = (pos_a_y_xyz-pos_a_o_xyz)/((
            numpy.square(pos_a_y_xyz-pos_a_o_xyz)).sum())**0.5
        v_pos_y_not_norm = v_pos_a_y - (v_pos_a_y * v_pos_x).sum() * v_pos_x
        v_pos_y = v_pos_y_not_norm / (
            numpy.square(v_pos_y_not_norm).sum())**0.5

        v_pos_z = numpy.cross(v_pos_x, v_pos_y)

        return v_pos_x, v_pos_y, v_pos_z

    def calc_fractions(self, cell: Cell, atom_site: AtomSiteL) -> \
            numpy.ndarray:
        """Give a numpy.nd_array of fractions: fractions_xyz.

        Arguments
        ---------
            - cell
            - atom_site
        """
        size_x, size_y = self.size_x, self.size_y
        points_x, points_y = self.points_x, self.points_y

        atom_o_label = self.atom_center
        _item_a_s_o = atom_site[atom_o_label]
        fract_a_o_xyz = numpy.array([_item_a_s_o.fract_x, _item_a_s_o.fract_y,
                                     _item_a_s_o.fract_z], dtype=float)
        pos_a_o_xyz = cell.calc_position_by_coordinate(
            fract_a_o_xyz[0], fract_a_o_xyz[1], fract_a_o_xyz[2])
        pos_a_o_xyz = numpy.array(pos_a_o_xyz, dtype=float)

        v_pos_x, v_pos_y, v_pos_z = self.calc_axes_x_y_z(cell, atom_site)

        v_delta_pos_x = v_pos_x * size_x / float(points_x)
        v_delta_pos_y = v_pos_y * size_y / float(points_y)
        np_x_2d, np_y_2d = numpy.meshgrid(range(-points_x//2, points_x//2),
                                          range(-points_y//2, points_y//2),
                                          indexing="ij")
        np_x_1d, np_y_1d = np_x_2d.flatten(), np_y_2d.flatten()

        pos_xyz = np_x_1d[numpy.newaxis, :]*v_delta_pos_x[:, numpy.newaxis] + \
            np_y_1d[numpy.newaxis, :] * v_delta_pos_y[:, numpy.newaxis] + \
            pos_a_o_xyz[:, numpy.newaxis]
        fract_x, fract_y, fract_z = cell.calc_coordinate_by_position(
            pos_xyz[0, :], pos_xyz[1, :], pos_xyz[2, :])
        return fract_x, fract_y, fract_z

    def calc_atoms(self, cell: Cell, atom_site: AtomSiteL,
                   space_group_symop: SpaceGroupSymopL, distance_min=0.3):
        """Calculate position of atoms in the section plane.

        Argument
        --------
            - cell
            - atom_site
            - space_group_symop
            - distance_min is minimal distance to search atom in angstrems
              (default value is 0.3)

        Output
        ------
            - atom_x
            - atom_y
            - atom_label
        """


        r_11 = numpy.array(space_group_symop.r_11, dtype=int)
        r_12 = numpy.array(space_group_symop.r_12, dtype=int)
        r_13 = numpy.array(space_group_symop.r_13, dtype=int)
        r_21 = numpy.array(space_group_symop.r_21, dtype=int)
        r_22 = numpy.array(space_group_symop.r_22, dtype=int)
        r_23 = numpy.array(space_group_symop.r_23, dtype=int)
        r_31 = numpy.array(space_group_symop.r_31, dtype=int)
        r_32 = numpy.array(space_group_symop.r_32, dtype=int)
        r_33 = numpy.array(space_group_symop.r_33, dtype=int)

        b_1 = numpy.array(space_group_symop.b_1, dtype=float)
        b_2 = numpy.array(space_group_symop.b_2, dtype=float)
        b_3 = numpy.array(space_group_symop.b_3, dtype=float)

        r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
        b_i = (b_1, b_2, b_3)

        fract_atom_auc_x = numpy.array(atom_site.fract_x, dtype=float)
        fract_atom_auc_y = numpy.array(atom_site.fract_y, dtype=float)
        fract_atom_auc_z = numpy.array(atom_site.fract_z, dtype=float)
        fract_atom_auc_xyz = (fract_atom_auc_x, fract_atom_auc_y,
                              fract_atom_auc_z)
        label_atom_auc = numpy.array(atom_site.label, dtype=str)

        fract_atom_uc_x, fract_atom_uc_y, fract_atom_uc_z, label_atom_uc = \
            calc_atoms_in_unit_cell(r_ij, b_i, fract_atom_auc_xyz,
                                    label_atom_auc)

        size_x = self.size_x
        size_y = self.size_y

        atom_center = self.atom_center
        atom_site_center = atom_site[atom_center]
        center_fract_x = atom_site_center.fract_x
        center_fract_y = atom_site_center.fract_y
        center_fract_z = atom_site_center.fract_z

        center_pos_x, center_pos_y, center_pos_z = \
            cell.calc_position_by_coordinate(center_fract_x, center_fract_y,
                                             center_fract_z)

        v_pos_x, v_pos_y, v_pos_z = self.calc_axes_x_y_z(cell, atom_site)

        pos_atom_uc_x, pos_atom_uc_y, pos_atom_uc_z = \
            cell.calc_position_by_coordinate(fract_atom_uc_x, fract_atom_uc_y,
                                             fract_atom_uc_z)

        pos_atom_loc_x = v_pos_x[0]*(pos_atom_uc_x - center_pos_x) + \
            v_pos_x[1]*(pos_atom_uc_y - center_pos_y) + \
            v_pos_x[2]*(pos_atom_uc_z - center_pos_z)
        pos_atom_loc_y = v_pos_y[0]*(pos_atom_uc_x - center_pos_x) + \
            v_pos_y[1]*(pos_atom_uc_y - center_pos_y) + \
            v_pos_y[2]*(pos_atom_uc_z - center_pos_z)
        pos_atom_loc_z = v_pos_z[0]*(pos_atom_uc_x - center_pos_x) + \
            v_pos_z[1]*(pos_atom_uc_y - center_pos_y) + \
            v_pos_z[2]*(pos_atom_uc_z - center_pos_z)

        flag_x = numpy.abs(pos_atom_loc_x) < 0.5*size_x
        flag_y = numpy.abs(pos_atom_loc_y) < 0.5*size_y
        flag_z = numpy.abs(pos_atom_loc_z) < distance_min
        flag_xyz = numpy.logical_and(flag_x, numpy.logical_and(flag_y, flag_z))

        atom_x = pos_atom_loc_x[flag_xyz]
        atom_y = pos_atom_loc_y[flag_xyz]
        atom_label = label_atom_uc[flag_xyz]
        return atom_x, atom_y, atom_label


class SectionL(LoopN):
    """Description of sections.

    Describe information concerning the density point.
    """

    ITEM_CLASS = Section
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SectionL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
# loop_
# _section_id
# _section_size_x
# _section_size_y
# _section_atom_center
# _section_atom_along_axis_x
# _section_atom_along_axis_x_symop
# _section_atom_along_axis_y
# _section_atom_along_axis_y_symop
# _section_points_x
# _section_points_y
# _section_url_out
#   core 5 5 Ho1 O1 x,y,z O2 x,y,z 100 100 s_ho1.dat
#   """

# obj = SectionL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["core"], end="\n\n")



import numpy


def calc_extinction_sphere_primary(
        f_sq, radius, volume_unit_cell, cos_2theta, wavelength,
        flag_f_sq: bool = False, flag_radius: bool = False,
        flag_volume_unit_cell: bool = False, flag_cos_2theta: bool = False,
        flag_wavelength: bool = False):
    """Primary extinction for the sphere (Acta.Cryst.(1974), A30, 129)

    """
    x = 1.5 * f_sq * numpy.square(wavelength*radius/volume_unit_cell)
    A = 0.20 + 0.45 * cos_2theta
    B = 0.22 - 0.12 * (0.5-cos_2theta)**2 
    x_sq = numpy.square(x)
    y_p = 1./numpy.sqrt(1 + 2 * x + A * x_sq / (1.+ B*x))
    dder = {}
    der_yp_x = -0.5*numpy.power(y_p, 3)*(
        2. + 2.*A*x / (1.+B*x) - A*B*x_sq / numpy.square(1.+ B*x))
    if flag_f_sq:
        ones_f_sq = numpy.ones_like(f_sq)
        dder["f_sq"] = der_yp_x * 1.5 * numpy.square(
            wavelength*radius/volume_unit_cell)*ones_f_sq
    if flag_radius:
        ones_radius = numpy.ones_like(radius)
        dder["radius"] = der_yp_x * 3.0 * f_sq * radius*numpy.square(
            wavelength/volume_unit_cell) 

    if flag_volume_unit_cell:
        ones_volume_unit_cell = numpy.ones_like(volume_unit_cell)
        dder["volume_unit_cell"] = -3.0 * der_yp_x *  f_sq * numpy.square(
            wavelength*radius/volume_unit_cell) * \
            ones_volume_unit_cell / volume_unit_cell
        pass
    if flag_cos_2theta:
        ones_cos_2theta = numpy.ones_like(cos_2theta)
        der_A_cos_2theta = 0.45 
        der_B_cos_2theta = 0.24 * (0.5-cos_2theta) 
        der_yp_A = -0.5*numpy.power(y_p, 3)*x_sq / (1.+ B*x)
        der_yp_B = 0.5*numpy.power(y_p, 3)*A * x_sq*x / numpy.square(1.+ B*x)
        dder["cos_2theta"] = (der_yp_A*der_A_cos_2theta +
                              der_yp_B*der_B_cos_2theta) * ones_cos_2theta
    if flag_wavelength:
        ones_wavelength = numpy.ones_like(wavelength)
        dder["wavelength"] = der_yp_x * 3.0 * f_sq * wavelength*numpy.square(
            radius/volume_unit_cell) * ones_wavelength
    return y_p, dder


def calc_extinction_sphere_secondary_gauss(
        f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
        flag_f_sq: bool = False, flag_radius: bool = False,
        flag_mosaicity: bool = False,
        flag_volume_unit_cell: bool = False, flag_cos_2theta: bool = False,
        flag_wavelength: bool = False):
    """Secondary Gauss extinction for the sphere (Acta.Cryst.(1974), A30, 129)
    mosaicity are given in minutes
    """
    mosaicity_rad = mosaicity*numpy.pi/(180*60) #transfer minutes to radians
    if not(radius*mosaicity_rad>0):
        term_1 = numpy.zeros_like(cos_2theta)
        x = numpy.zeros_like(term_1)
    else:
        term_1 = mosaicity_rad / numpy.sqrt(numpy.square(mosaicity_rad * wavelength) +
                                            9./8. * numpy.square(radius) *
                                            (1.- numpy.square(cos_2theta)))
        x = 1.5 * f_sq * wavelength * \
            numpy.square(wavelength*radius/volume_unit_cell) * term_1
    A = 0.58 + 0.48 * cos_2theta + 0.24 * numpy.square(cos_2theta)
    B = 0.02 - 0.025 * cos_2theta
    x_sq = numpy.square(x)
    y_s = 1./numpy.sqrt(1 + 2.12 * x + A * x_sq / (1.+ B*x))
    y_s[numpy.isnan(y_s)]=1. 
    dder = {}
    der_ys_x = -0.5*numpy.power(y_s, 3)*(
        2.12 + 2.*A*x / (1.+B*x) - A*B*x_sq / numpy.square(1.+ B*x))
    if flag_f_sq:
        ones_f_sq = numpy.ones_like(f_sq)
        dder["f_sq"] = der_ys_x * 1.5 * wavelength*numpy.square(
            wavelength*radius/volume_unit_cell)*term_1*ones_f_sq
    if flag_radius:
        ones_radius = numpy.ones_like(radius)
        dder["radius"] = der_ys_x*(2*radius*1.5 * f_sq * wavelength * \
            numpy.square(wavelength/volume_unit_cell)*term_1+
            1.5 * f_sq * wavelength * \
            numpy.square(wavelength*radius/volume_unit_cell) * mosaicity_rad / \
            numpy.power(numpy.sqrt(numpy.square(mosaicity_rad * wavelength) + \
            9./8. * numpy.square(radius) * (1.- numpy.square(cos_2theta))),3) *
            9./8. * 2*radius * (1.- numpy.square(cos_2theta))
            )
    if flag_mosaicity:
        dder["mosaicity"] = numpy.pi/(180*60)*der_ys_x*1.5 * f_sq * wavelength * \
            numpy.square(wavelength*radius/volume_unit_cell)*(
                 1./ numpy.sqrt(numpy.square(mosaicity_rad * wavelength) +
                    9./8. * numpy.square(radius) *(1.- numpy.square(cos_2theta))) + 
                                            mosaicity_rad / \
            numpy.power(numpy.sqrt(numpy.square(mosaicity_rad * wavelength) + \
            9./8. * numpy.square(radius) * (1.- numpy.square(cos_2theta))),3)*2.*numpy.square(wavelength)*mosaicity_rad)
    if flag_volume_unit_cell:
        ones_volume_unit_cell = numpy.ones_like(volume_unit_cell)
        dder["volume_unit_cell"] = 0*ones_volume_unit_cell
    if flag_cos_2theta:
        ones_cos_2theta = numpy.ones_like(cos_2theta)
        dder["cos_2theta"] = 0*ones_cos_2theta
    if flag_wavelength:
        ones_wavelength = numpy.ones_like(wavelength)
        dder["wavelength"] = 0*ones_wavelength
    return y_s, dder


def calc_extinction_sphere_secondary_lorentz(
        f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
        flag_f_sq: bool = False, flag_radius: bool = False,
        flag_mosaicity: bool = False,
        flag_volume_unit_cell: bool = False, flag_cos_2theta: bool = False,
        flag_wavelength: bool = False):
    """Secondary Gauss extinction for the sphere (Acta.Cryst.(1974), A30, 129)
    mosaicity are given in minutes
    """
    mosaicity_rad = mosaicity*numpy.pi/(180*60) #transfer minutes to radians
    if not(radius*mosaicity_rad>0):
        term_1 = numpy.zeros_like(cos_2theta)
        x = numpy.zeros_like(term_1)
    else:
        term_1 = mosaicity_rad / (mosaicity_rad * wavelength + radius * numpy.sqrt(
            1.- numpy.square(cos_2theta)))
        x = 1.5 * f_sq * wavelength * \
            numpy.square(wavelength*radius/volume_unit_cell) * term_1
    A = 0.025 + 0.285 * cos_2theta
    B_1 = -0.45 * cos_2theta # cos_2theta <= 0.
    B_2 = 0.15 - 0.2 * numpy.square(0.75 - cos_2theta) # cos_2theta > 0.
    B = numpy.where(cos_2theta> 0., B_2, B_1)
    x_sq = numpy.square(x)
    y_s = 1./numpy.sqrt(1 + 2 * x + A * x_sq / (1.+ B*x))
    dder = {}
    der_ys_x = -0.5*numpy.power(y_s, 3)*(
        2. + 2.*A*x / (1.+B*x) - A*B*x_sq / numpy.square(1.+ B*x))
    if flag_f_sq:
        ones_f_sq = numpy.ones_like(f_sq)
        dder["f_sq"] = der_ys_x * 1.5 * wavelength*numpy.square(
            wavelength*radius/volume_unit_cell)*term_1*ones_f_sq
    if flag_radius:
        ones_radius = numpy.ones_like(radius)
        dder["radius"] = 0*ones_radius
    if flag_mosaicity:
        ones_mosaicity = numpy.ones_like(mosaicity)
        dder["mosaicity"] = 0*ones_mosaicity
    if flag_volume_unit_cell:
        ones_volume_unit_cell = numpy.ones_like(volume_unit_cell)
        dder["volume_unit_cell"] = 0*ones_volume_unit_cell
    if flag_cos_2theta:
        ones_cos_2theta = numpy.ones_like(cos_2theta)
        dder["cos_2theta"] = 0*ones_cos_2theta
    if flag_wavelength:
        ones_wavelength = numpy.ones_like(wavelength)
        dder["wavelength"] = 0*ones_wavelength
    return y_s, dder


def calc_extinction_sphere(
        f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
        model: str, flag_f_sq: bool = False, flag_radius: bool = False,
        flag_mosaicity: bool = False,
        flag_volume_unit_cell: bool = False, flag_cos_2theta: bool = False,
        flag_wavelength: bool = False):

    """Extinction correction for the sphere (Acta.Cryst.(1974), A30, 129)
    
    model: "gauss" or "lorentz"
    mosaicity are given in minutes
    """
    y_p, dder_y_p = calc_extinction_sphere_primary(
        f_sq, radius, volume_unit_cell, cos_2theta, wavelength,
        flag_f_sq=flag_f_sq, flag_radius=flag_radius,
        flag_volume_unit_cell=flag_volume_unit_cell,
        flag_cos_2theta=flag_cos_2theta, flag_wavelength=flag_wavelength)
    if model.lower().startswith("gauss"):
        y_s, dder_y_s = calc_extinction_sphere_secondary_gauss(
            f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
            flag_f_sq=flag_f_sq, flag_radius=flag_radius,
            flag_mosaicity=flag_mosaicity,
            flag_volume_unit_cell=flag_volume_unit_cell,
            flag_cos_2theta=flag_cos_2theta, flag_wavelength=flag_wavelength)
    elif model.lower().startswith("lorentz"):
        y_s, dder_y_s = calc_extinction_sphere_secondary_lorentz(
            f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
            flag_f_sq=flag_f_sq, flag_radius=flag_radius,
            flag_mosaicity=flag_mosaicity,
            flag_volume_unit_cell=flag_volume_unit_cell,
            flag_cos_2theta=flag_cos_2theta, flag_wavelength=flag_wavelength)
    else:
        y_s = numpy.ones(y_p.shape, dtype=float)
        dder_y_s = {
            "f_sq": numpy.zeros_like(f_sq),
            "radius": numpy.zeros_like(radius),
            "mosaicity": numpy.zeros_like(mosaicity),
            "volume_unit_cell": numpy.zeros_like(volume_unit_cell),
            "cos_2theta": numpy.zeros_like(cos_2theta),
            "wavelength": numpy.zeros_like(wavelength)}
    y = y_p * y_s
    dder = {}
    if flag_f_sq:
        dder["f_sq"] = y_p * dder_y_s["f_sq"] + dder_y_p["f_sq"] * y_s
    if flag_radius:
        dder["radius"] = y_p * dder_y_s["radius"] + dder_y_p["radius"] * y_s
    if flag_mosaicity:
        dder["mosaicity"] = y_p * dder_y_s["mosaicity"] 
    if flag_volume_unit_cell:
        dder["volume_unit_cell"] = y_p * dder_y_s["volume_unit_cell"] + \
            dder_y_p["volume_unit_cell"] * y_s
    if flag_cos_2theta:
        dder["cos_2theta"] = y_p * dder_y_s["cos_2theta"] + \
            dder_y_p["cos_2theta"] * y_s
    if flag_wavelength:
        dder["wavelength"] = y_p * dder_y_s["wavelength"] + \
            dder_y_p["wavelength"] * y_s
    return y, dder



"""MEMParameters, MEMParametersL classes."""
from typing import NoReturn

import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class MEMParameters(ItemN):
    """Description of MEMParameters.

    Attributes
    ----------
        - points_a, points_b, points_c, chi_ferro (mandatory)
        - chi_antiferro, prior_density, method (mandatory)
        - gof_desired (mandatory)

    Accessible Values
    -----------------
        method is "2channel" (default) or "tensorMEM"
        prior_density is "uniform" (default) or "core"
    """

    ATTR_MANDATORY_NAMES = ("points_a", "points_b", "points_c", )
    ATTR_MANDATORY_TYPES = (int, int, int, )
    ATTR_MANDATORY_CIF = ("points_a", "points_b", "points_c", )

    ATTR_OPTIONAL_NAMES = ("channel_plus_minus",
                          "channel_chi", "magnetization_plus", "magnetization_minus",
                          "prior_density", "use_asymmetry", "only_magnetic_basins",
                          "gof_desired", "spin_density_to_den_file", "magnetization_density_to_den_file")
    ATTR_OPTIONAL_TYPES = (bool, bool, float, float, str, bool, bool, float, str, str)
    ATTR_OPTIONAL_CIF = ("channel_plus_minus",
                          "channel_chi", "magnetization_plus", "magnetization_minus",
                          "prior_density", "use_asymmetry", "only_magnetic_basins",
                          "GoF_desired", "spin_density_to_den_file", "magnetization_density_to_den_file")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("magnetization_plus", "magnetization_minus")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'magnetization_plus': "{:.5f}", 'magnetization_minus': "{:.5f}",
                 'gof_desired': "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"prior_density": ["core", "uniform"],
                     }

    # default values for the parameters
    D_DEFAULT = {"points_a": 48, "points_b": 48, "points_c": 48,
                 "channel_plus_minus": True,
                 "magnetization_plus": 4., "magnetization_minus": -1.,
                 "channel_chi": False,
                 "only_magnetic_basins": False,
                 "prior_density": "uniform", "use_asymmetry": True,
                 "gof_desired": 1.,
                 "spin_density_to_den_file": "mem_spin_density.den",
                 "magnetization_density_to_den_file": "mem_magnetization_density.den",}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "mem_parameters"

    def __init__(self, **kwargs) -> NoReturn:
        super(MEMParameters, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"points_a": 1, "points_b": 1, "points_c": 1, "magnetization_plus": 0.,
                 "gof_desired": 0.}

        # defined for ani integer and float parameters
        D_MAX = {"magnetization_minus": 0}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_dictionary(self):
        numb_a = self.points_a
        numb_b = self.points_b
        numb_c = self.points_c
        points_abc = numpy.array([numb_a, numb_b, numb_c], dtype=int)
        
        channel_plus_minus = self.channel_plus_minus
        channel_chi = self.channel_chi
        magnetization_plus = self.magnetization_plus
        magnetization_minus = self.magnetization_minus
        flag_uniform_prior_density = self.prior_density == "uniform"
        flag_asymmetry = self.use_asymmetry
        flag_only_magnetic_basins = self.only_magnetic_basins
        gof_desired = self.gof_desired

        file_spin_density = self.spin_density_to_den_file
        file_magnetization_density = self.magnetization_density_to_den_file
        dict_mem = {
            'points_abc': points_abc,
            "channel_plus_minus": channel_plus_minus,
            "channel_chi": channel_chi,
            "magnetization_plus": magnetization_plus,
            "magnetization_minus": magnetization_minus,
            "flag_uniform_prior_density": flag_uniform_prior_density,
            "flag_asymmetry": flag_asymmetry,
            "flag_only_magnetic_basins": flag_only_magnetic_basins,
            "gof_desired": gof_desired,
            "file_spin_density": file_spin_density,
            "file_magnetization_density": file_magnetization_density,
            }
        return dict_mem

class MEMParametersL(LoopN):
    """Description of chi2 in loop."""

    ITEM_CLASS = MEMParameters
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(MEMParametersL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

